#  Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Train preprocessor helps doing data pre-processing during training.
"""
import logging
from typing import Optional, Dict, Type, Any, Union, Iterator, List


from forte.common.configuration import Config
from forte.data.base_extractor import BaseExtractor
from forte.data.converter import Converter
from forte.data.data_pack import DataPack
from forte.data.data_pack_dataset import DataPackDataset, DataPackIterator
from forte.data.ontology import Annotation
from forte.data.ontology.core import EntryType
from forte.utils import extractor_utils, create_import_error_msg
from forte.utils.extractor_utils import parse_feature_extractors

try:
    import torch
    from torch import device
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "torch", "extractor", "the extrator system support"
        )
    ) from e

try:
    from texar.torch.data import DataIterator, Batch
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "extractor", "the extrator system support"
        )
    ) from e


logger = logging.getLogger(__name__)

__all__ = ["TrainPreprocessor"]


class TrainPreprocessor:
    r"""
    `TrainPreprocessor` provides the functionality of doing pre-processing work
    including building vocabulary, extracting the features, batching and
    padding (optional). The processed data will be provided by its method
    :meth:`get_train_batch_iterator`, which will return an `iterator` over the
    batch of pre-processed data. Please refer to the documentation of
    that method for how the pre-processing is done.

    A main part of the `TrainPreprocessor ` is that it maintains a list of
    extractors :class:`~forte.data.base_extractor.BaseExtractor` that extract features. This
    can be provided either via calling `add_extractor` function. Alternatively,
    a request can be passed in through `initialize`, where the configuration
    under the `request` key will be used to create the extractor instances.

    The parsed components will be stored, and can be accessed via the `request`
    property of this class.

    Args:
        pack_iterator (Iterator[DataPack]): An iterator of
            :class:`~forte.data.data_pack.DataPack`.

    .. note::

        For parameters `request`, user does not necessarily need to provide
        `converter`. If no `converter` is specified, a default converter of
        type :class:`~forte.data.converter.converter.Converter` will be picked.
    """

    DATA_INPUT = extractor_utils.DATA_INPUT
    DATA_OUTPUT = extractor_utils.DATA_OUTPUT

    def __init__(self, pack_iterator: Iterator[DataPack]):
        self._pack_iterator: Iterator[DataPack] = pack_iterator
        self._cached_packs: List[DataPack] = []

        self._config: Config = None
        self._user_request: Dict = {}
        # Parsed feature extractors.
        self._request: Dict = {}
        self._request_ready: bool = False
        self._vocab_ready: bool = False

    def initialize(self, config: Optional[Union[Config, Dict]] = None):
        self._config = Config(
            config,
            default_hparams=self.default_configs(),
            allow_new_hparam=True,
        )
        self._user_request = self._config.request
        self._validate_config()
        self._parse_request(self._user_request)
        self._build_vocab()

    def add_extractor(
        self,
        name: str,
        extractor: BaseExtractor,
        is_input: bool,
        converter: Optional[Converter] = None,
    ):
        """
        Extractors can be added to the preprocessor directly via this
        method.

        Args:
            name: The name/identifier of this extractor, the name should be
              different between different extractors.
            extractor: The extractor instance to be added.
            is_input: Whether this extractor will be used as input or output.
            converter:  The converter instance to be applied after running
              the extractor.

        Returns:

        """
        extractor_utils.add_extractor(
            self._request, name, extractor, is_input, converter
        )

    @staticmethod
    def default_configs():
        r"""Returns a dictionary of default hyper-parameters.

        .. code-block:: python

            {
                "preprocess": {
                            "device": "cpu",
                },
                "dataset": DataPackDataset.default_hparams()
            }

        Here:

            - `"preprocessor.device"`:
              The device of the produced batches. For GPU training,
              set to current CUDA device.

            - `"dataset"`:
              This contains all the configurable options same as
              :class:`~forte.data.data_pack_dataset.DataPackDataset`.
        """

        # Configs should be serializable
        return {
            "preprocess": {
                "device": "cpu",
            },
            "dataset": DataPackDataset.default_hparams(),
            "request": {"context_type": None, "feature_scheme": None},
        }

    def _validate_config(self):
        # Placeholder
        pass

    def _parse_request(self, request: Dict):
        """
        This method has two responsibilities:
        1. parse the given data request and stored it internally
        2. validate if the given data request is valid
        """
        parsed_request: Dict[str, Any] = {}

        if "context_type" not in request or request["context_type"] is None:
            raise ValueError("Field not found for data request: `context_type`")

        if "feature_scheme" not in request or request["feature_scheme"] is None:
            raise ValueError(
                "Field not found for data request: `feature_scheme`"
            )

        parsed_request["context_type"] = request["context_type"]
        parsed_request["schemes"] = parse_feature_extractors(
            request["feature_scheme"]
        )

        self._request = parsed_request
        self._request_ready = True

    def _build_vocab(self):
        context_type: EntryType = self._request["context_type"]
        schemes: Dict = self._request["schemes"]

        # TODO: clear vocab?

        # Cached all data packs
        # TODO: this caching is not scalable
        for data_pack in self._pack_iterator:
            self._cached_packs.append(data_pack)

        for _, scheme in schemes.items():
            extractor: BaseExtractor = scheme["extractor"]
            if extractor.vocab_method != "raw":
                for data_pack in self._cached_packs:
                    if context_type is None:
                        extractor.update_vocab(data_pack)
                    else:
                        context: Annotation
                        for context in data_pack.get(context_type):
                            extractor.update_vocab(data_pack, context)

        self._vocab_ready = True

    def _build_dataset_iterator(self) -> DataIterator:
        context_type: Type[EntryType] = self._request["context_type"]  # type: ignore
        schemes: Dict[str, Dict[str, Any]] = self._request["schemes"]

        data_source = DataPackIterator(
            pack_iterator=iter(self._cached_packs),
            context_type=context_type,
            request={context_type: []},
        )

        dataset = DataPackDataset(
            data_source, schemes, self._config.dataset, self.device
        )
        iterator = DataIterator(dataset)

        return iterator

    @property
    def request(self) -> Dict:
        # pylint: disable=line-too-long
        r"""
        A `Dict` containing all the information needed for doing the
        pre-processing. This is obtained via parsing the input `request`

            An example `request` is:

            .. code-block:: python

                request = {
                    "context_type": "ft.onto.base_ontology.Sentence"
                    "schemes": {
                        "text_tag": {
                            "extractor":
                                "class_name":
                                  "forte.data.extractor.AttributeExtractor",
                                "config": {
                                    ... more configuration of the extractor
                                }
                        },
                        "ner_tag": {
                            "extractor":
                                "class_name":
                                  "forte.data.extractor.BioSeqTaggingExtractor",
                                "config": {
                                    ... more configuration of the extractor
                                }
                        }
                    }
                }

        Here:

            - `"context_type"`: Annotation
              A class of type :class:`~ft.onto.base_ontology.context_type`.
              Defines the granularity to separate data into different
              groups. All extractors will operate based on this. For example,
              if `context_type` is :class:`~ft.onto.base_ontology.Sentence`,
              then the features of each extractor will represent the
              information of a sentence. If this value is `None`, then all
              extractors will operate on the whole data pack.

            - `"schemes"`: Dict
              A Dict containing the information about doing the
              pre-processing.
              The `key` is the tags provided by input `request`. The
              `value` is a `Dict` containing the information for doing
              pre-processing for that feature.

            - `"schemes.tag.extractor"`:
              An instance of type
              :class:`~forte.data.extractor.BaseExtractor`.

            - `"schemes.tag.converter"`:
              An instance of type
              :class:`~forte.data.converter.converter.Converter`.

            - `"schemes.tag.type"`: TrainPreprocessor.DATA_INPUT/DATA_OUTPUT
              Denoting whether this feature is the input or output feature.
        """
        if not self._request:
            self._parse_request(self._request)
        return self._request

    @property
    def device(self) -> device:
        r"""The device of the produced batches. For GPU training,
        set to current CUDA device.
        """
        return torch.device(self._config.preprocess.device)

    @property
    def config(self) -> Config:
        r"""A :class:`~forte.common.configuration.Config` maintaining all the
        configurable options for this `TrainPreprocessor`.
        """
        return self._config

    def get_train_batch_iterator(self) -> Iterator[Batch]:
        r"""
        This method mainly has four steps:

        1. Iterate over :class:`~forte.data.data_pack.DataPack`
           via pack iterator
        2. Extract :class:`~forte.data.converter.feature.Feature` from
           :class:`~forte.data.data_pack.DataPack`
        3. Batch :class:`~forte.data.converter.feature.Feature`
        4. (optional) Pad a batch of
           :class:`~forte.data.converter.feature.Feature`

        It will return an `iterator` of a batch of pre-processed data.

        Returns:
            An `Iterator` of type :class:`~texar.torch.data.Batch`

            Please refer to :meth:`collate` in
            :class:`~forte.data.data_pack_dataset.DataPackDataset` for details
            about its structure.
        """
        if not self._request:
            raise ValueError("Feature resource is not parsed")

        if not self._vocab_ready:
            raise ValueError("Vocab is not built")

        dataset_iter = self._build_dataset_iterator()

        return iter(dataset_iter)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Processors that process pack.
"""
from abc import ABC

from forte.data.base_pack import PackType
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.processors.base.base_processor import BaseProcessor

__all__ = ["BasePackProcessor", "PackProcessor", "MultiPackProcessor"]


class BasePackProcessor(BaseProcessor[PackType], ABC):
    r"""The base class of processors that process one pack in a streaming way.
    If you are looking for batching (that might happen across packs, refer to
    :class:`~forte.processors.base.batch_processor.BaseBatchProcessor`.
    """
    pass


class PackProcessor(BasePackProcessor[DataPack], ABC):
    r"""The base class of processors that process one :class:`~forte.data.data_pack.DataPack` each
    time.
    """

    def _process(self, input_pack: DataPack):
        raise NotImplementedError


class MultiPackProcessor(BasePackProcessor[MultiPack], ABC):
    r"""
    The base class of processors that process :class:`~forte.data.multi_pack.MultiPack` each time.
    """

    def _process(self, input_pack: MultiPack):
        raise NotImplementedError



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various utilities specific to data processing.
"""
import logging
import os
import sys
import tarfile
import urllib.request
import zipfile
import re
from typing import List, Optional, overload, Union

from forte.utils.types import PathLike
from forte.utils.utils_io import maybe_create_dir

__all__ = [
    "maybe_download",
]


# TODO: Remove these once pylint supports function stubs.
# pylint: disable=unused-argument,function-redefined,missing-docstring


@overload
def maybe_download(
    urls: List[str],
    path: Union[str, PathLike],
    filenames: Optional[List[str]] = None,
    extract: bool = False,
    num_gdrive_retries: int = 1,
) -> List[str]:
    ...


@overload
def maybe_download(
    urls: str,
    path: Union[str, PathLike],
    filenames: Optional[str] = None,
    extract: bool = False,
    num_gdrive_retries: int = 1,
) -> str:
    ...


def maybe_download(
    urls: Union[List[str], str],
    path: Union[str, PathLike],
    filenames: Union[List[str], str, None] = None,
    extract: bool = False,
    num_gdrive_retries: int = 1,
):
    r"""Downloads a set of files.

    Args:
        urls: A (list of) URLs to download files.
        path: The destination path to save the files.
        filenames: A (list of) strings of the file names. If given,
            must have the same length with ``urls``. If `None`,
            filenames are extracted from ``urls``.
        extract: Whether to extract compressed files.
        num_gdrive_retries: An integer specifying the number of attempts
            to download file from Google Drive. Default value is 1.

    Returns:
        A list of paths to the downloaded files.
    """
    maybe_create_dir(path)

    if not isinstance(urls, (list, tuple)):
        is_list = False
        urls = [urls]
    else:
        is_list = True
    if filenames is not None:
        if not isinstance(filenames, (list, tuple)):
            filenames = [filenames]
        if len(urls) != len(filenames):
            raise ValueError(
                "`filenames` must have the same number of elements as `urls`."
            )

    result = []
    for i, url in enumerate(urls):
        if filenames is not None:
            filename = filenames[i]
        elif "drive.google.com" in url:
            filename = _extract_google_drive_file_id(url)
        else:
            filename = url.split("/")[-1]
            # If downloading from GitHub, remove suffix ?raw=True
            # from local filename
            if filename.endswith("?raw=true"):
                filename = filename[:-9]

        filepath = os.path.join(path, filename)
        result.append(filepath)

        # if not tf.gfile.Exists(filepath):
        if not os.path.exists(filepath):
            if "drive.google.com" in url:
                filepath = _download_from_google_drive(
                    url, filename, path, num_gdrive_retries
                )
            else:
                filepath = _download(url, filename, path)

            if extract:
                logging.info("Extract %s", filepath)
                if tarfile.is_tarfile(filepath):
                    with tarfile.open(filepath, "r") as tfile:
                        tfile.extractall(path)
                elif zipfile.is_zipfile(filepath):
                    with zipfile.ZipFile(filepath) as zfile:
                        zfile.extractall(path)
                else:
                    logging.info(
                        "Unknown compression type. Only .tar.gz"
                        ".tar.bz2, .tar, and .zip are supported"
                    )
    if not is_list:
        return result[0]
    return result


# pylint: enable=unused-argument,function-redefined,missing-docstring


def _download(url: str, filename: str, path: Union[PathLike, str]) -> str:
    def _progress_hook(count, block_size, total_size):
        percent = float(count * block_size) / float(total_size) * 100.0
        sys.stdout.write(f"\r>> Downloading {filename} {percent:.1f}%")
        sys.stdout.flush()

    filepath = os.path.join(path, filename)
    filepath, _ = urllib.request.urlretrieve(url, filepath, _progress_hook)
    print()
    statinfo = os.stat(filepath)
    logging.info(
        "Successfully downloaded %s %d bytes", filename, statinfo.st_size
    )

    return filepath


def _extract_google_drive_file_id(url: str) -> str:
    # id is between `/d/` and '/'
    url_suffix = url[url.find("/d/") + 3 :]
    if url_suffix.find("/") == -1:
        # if there's no trailing '/'
        return url_suffix
    file_id = url_suffix[: url_suffix.find("/")]
    return file_id


def _download_from_google_drive(
    url: str, filename: str, path: Union[str, PathLike], num_retries: int = 1
) -> str:
    r"""Adapted from `https://github.com/saurabhshri/gdrive-downloader`"""

    # pylint: disable=import-outside-toplevel
    try:
        import requests
        from requests import HTTPError
    except ImportError:
        logging.info(
            "The requests library must be installed to download files from "
            "Google drive. Please see: https://github.com/psf/requests"
        )
        raise

    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith("download_warning"):
                return value
        if "Google Drive - Virus scan warning" in response.text:
            match = re.search("confirm=([0-9A-Za-z_]+)", response.text)
            if match is None or len(match.groups()) < 1:
                raise ValueError(
                    "No token found in warning page from Google Drive."
                )
            return match.groups()[0]
        return None

    file_id = _extract_google_drive_file_id(url)

    gurl = "https://docs.google.com/uc?export=download"
    sess = requests.Session()
    params = {"id": file_id}
    response = sess.get(gurl, params=params, stream=True)
    token = _get_confirm_token(response)

    if token:
        params = {"id": file_id, "confirm": token}
        response = sess.get(gurl, params=params, stream=True)
    while response.status_code != 200 and num_retries > 0:
        response = requests.get(gurl, params=params, stream=True)
        num_retries -= 1
    if response.status_code != 200:
        logging.error(
            "Failed to download %s because of invalid response "
            "from %s: status_code='%d' reason='%s' content=%s",
            filename,
            response.url,
            response.status_code,
            response.reason,
            response.content,
        )
        raise HTTPError(response=response)

    filepath = os.path.join(path, filename)
    CHUNK_SIZE = 32768
    with open(filepath, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:
                f.write(chunk)

    logging.info("Successfully downloaded %s", filename)

    return filepath



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Conditional random field.
Adapted from AllenNLP but removed the feature of external restriction
"""
__all__ = ["ConditionalRandomField"]
from typing import Optional, List, Tuple, Dict, Union
import logging
from forte.utils import create_import_error_msg

try:
    import torch
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "conditional random fields")
    ) from e

logger = logging.getLogger(__name__)


def allowed_transitions(
    constraint_type: str, labels: Dict[int, str]
) -> List[Tuple[int, int]]:
    """
    Given labels and a constraint type, returns the allowed transitions. It will
    additionally include transitions for the start and end states, which are
    used by the conditional random field.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    labels : ``Dict[int, str]``, required
        A mapping {label_id -> label}. Most commonly this would be the value
        from Vocabulary.get_index_to_token_vocabulary()

    Returns
    -------
    ``List[Tuple[int, int]]``
        The allowed transitions (from_label_id, to_label_id).
    """
    num_labels = len(labels)
    start_tag = num_labels
    end_tag = num_labels + 1
    labels_with_boundaries = list(labels.items()) + [
        (start_tag, "START"),
        (end_tag, "END"),
    ]

    allowed = []
    for from_label_index, from_label in labels_with_boundaries:
        if from_label in ("START", "END"):
            from_tag = from_label
            from_entity = ""
        else:
            from_tag = from_label[0]
            from_entity = from_label[1:]
        for to_label_index, to_label in labels_with_boundaries:
            if to_label in ("START", "END"):
                to_tag = to_label
                to_entity = ""
            else:
                to_tag = to_label[0]
                to_entity = to_label[1:]
            if is_transition_allowed(
                constraint_type, from_tag, from_entity, to_tag, to_entity
            ):
                allowed.append((from_label_index, to_label_index))
    return allowed


def is_transition_allowed(
    constraint_type: str,
    from_tag: str,
    from_entity: str,
    to_tag: str,
    to_entity: str,
):
    """
    Given a constraint type and strings ``from_tag`` and ``to_tag`` that
    represent the origin and destination of the transition, return whether
    the transition is allowed under the given constraint type.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    from_tag : ``str``, required
        The tag that the transition originates from. For example, if the
        label is ``I-PER``, the ``from_tag`` is ``I``.
    from_entity: ``str``, required
        The entity corresponding to the ``from_tag``. For example, if the
        label is ``I-PER``, the ``from_entity`` is ``PER``.
    to_tag : ``str``, required
        The tag that the transition leads to. For example, if the
        label is ``I-PER``, the ``to_tag`` is ``I``.
    to_entity: ``str``, required
        The entity corresponding to the ``to_tag``. For example, if the
        label is ``I-PER``, the ``to_entity`` is ``PER``.

    Returns
    -------
    ``bool``
        Whether the transition is allowed under the given ``constraint_type``.
    """
    # pylint: disable=too-many-return-statements
    if to_tag == "START" or from_tag == "END":
        # Cannot transition into START or from END
        return False

    if constraint_type == "BIOUL":
        if from_tag == "START":
            return to_tag in ("O", "B", "U")
        if to_tag == "END":
            return from_tag in ("O", "L", "U")
        return any(
            [
                # O can transition to O, B-* or U-*
                # L-x can transition to O, B-*, or U-*
                # U-x can transition to O, B-*, or U-*
                from_tag in ("O", "L", "U") and to_tag in ("O", "B", "U"),
                # B-x can only transition to I-x or L-x
                # I-x can only transition to I-x or L-x
                from_tag in ("B", "I")
                and to_tag in ("I", "L")
                and from_entity == to_entity,
            ]
        )
    elif constraint_type == "BIO":
        if from_tag == "START":
            return to_tag in ("O", "B")
        if to_tag == "END":
            return from_tag in ("O", "B", "I")
        return any(
            [
                # Can always transition to O or B-x
                to_tag in ("O", "B"),
                # Can only transition to I-x from B-x or I-x
                to_tag == "I"
                and from_tag in ("B", "I")
                and from_entity == to_entity,
            ]
        )
    elif constraint_type == "IOB1":
        if from_tag == "START":
            return to_tag in ("O", "I")
        if to_tag == "END":
            return from_tag in ("O", "B", "I")
        return any(
            [
                # Can always transition to O or I-x
                to_tag in ("O", "I"),
                # Can only transition to B-x from B-x or I-x, where
                # x is the same tag.
                to_tag == "B"
                and from_tag in ("B", "I")
                and from_entity == to_entity,
            ]
        )
    elif constraint_type == "BMES":
        if from_tag == "START":
            return to_tag in ("B", "S")
        if to_tag == "END":
            return from_tag in ("E", "S")
        return any(
            [
                # Can only transition to B or S from E or S.
                to_tag in ("B", "S") and from_tag in ("E", "S"),
                # Can only transition to M-x from B-x, where
                # x is the same tag.
                to_tag == "M"
                and from_tag in ("B", "M")
                and from_entity == to_entity,
                # Can only transition to E-x from B-x or M-x, where
                # x is the same tag.
                to_tag == "E"
                and from_tag in ("B", "M")
                and from_entity == to_entity,
            ]
        )
    else:
        raise ValueError(f"Unknown constraint type: {constraint_type}")


class ConditionalRandomField(torch.nn.Module):
    """
    This module uses the "forward-backward" algorithm to compute
    the log-likelihood of its inputs assuming a conditional random field model.

    See, e.g. http://www.cs.columbia.edu/~mcollins/fb.pdf

    Args:
        num_tags: The number of tags.
        constraints:
            An optional list of allowed transitions (from_tag_id, to_tag_id).
            These are applied to ``viterbi_tags()`` but do not affect
            ``forward()``. These should be derived from `allowed_transitions`
            so that the start and end transitions are handled correctly for
            your tag type.
        include_start_end_transitions:
            Whether to include the start and end transition parameters.
    """

    def __init__(
        self,
        num_tags: int,
        constraints: Optional[List[Tuple[int, int]]] = None,
        include_start_end_transitions: bool = True,
    ) -> None:
        super().__init__()
        self.num_tags = num_tags

        # transitions[i, j] is the logit for transitioning from state i
        # to state j.
        self.transitions = torch.nn.Parameter(torch.Tensor(num_tags, num_tags))

        # _constraint_mask indicates valid transitions (based on supplied
        # constraints).
        # Include special start of sequence (num_tags + 1) and end of sequence
        # tags (num_tags + 2)
        if constraints is None:
            # All transitions are valid.
            constraint_mask = torch.Tensor(num_tags + 2, num_tags + 2).fill_(
                1.0
            )
        else:
            constraint_mask = torch.Tensor(num_tags + 2, num_tags + 2).fill_(
                0.0
            )
            for i, j in constraints:
                constraint_mask[i, j] = 1.0

        self._constraint_mask = torch.nn.Parameter(
            constraint_mask, requires_grad=False
        )

        # Also need logits for transitioning from "start" state and to "end"
        # state.
        self.include_start_end_transitions = include_start_end_transitions
        if include_start_end_transitions:
            self.start_transitions = torch.nn.Parameter(torch.Tensor(num_tags))
            self.end_transitions = torch.nn.Parameter(torch.Tensor(num_tags))

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_normal_(self.transitions)
        if self.include_start_end_transitions:
            torch.nn.init.normal_(self.start_transitions)
            torch.nn.init.normal_(self.end_transitions)

    def _input_likelihood(
        self, logits: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """
        Computes the (batch_size,) denominator term for the log-likelihood,
        which is the um of the likelihoods across all possible state sequences.
        """
        batch_size, sequence_length, num_tags = logits.size()

        # Transpose batch size and sequence dimensions
        mask = mask.float().transpose(0, 1).contiguous()
        logits = logits.transpose(0, 1).contiguous()

        # Initial alpha is the (batch_size, num_tags) tensor of likelihoods
        # combining the transitions to the initial states and the logits for
        # the first timestep.
        if self.include_start_end_transitions:
            alpha = self.start_transitions.view(1, num_tags) + logits[0]
        else:
            alpha = logits[0]

        # For each i we compute logits for the transitions from timestep i-1
        # to timestep i.
        # We do so in a (batch_size, num_tags, num_tags) tensor where the axes
        # are (instance, current_tag, next_tag)
        for i in range(1, sequence_length):
            # The emit scores are for time i ("next_tag") so we broadcast along
            # the current_tag axis.
            emit_scores = logits[i].view(batch_size, 1, num_tags)
            # Transition scores are (current_tag, next_tag) so we broadcast
            # along the instance axis.
            transition_scores = self.transitions.view(1, num_tags, num_tags)
            # Alpha is for the current_tag, so we broadcast along the next
            # tag axis.
            broadcast_alpha = alpha.view(batch_size, num_tags, 1)

            # Add all the scores together and logexp over the current_tag axis
            inner = broadcast_alpha + emit_scores + transition_scores

            # In valid positions (mask == 1) we want to take the logsumexp over
            # the current_tag dimension
            # of ``inner``. Otherwise (mask == 0) we want to retain the
            # previous alpha.
            alpha = torch.logsumexp(inner, 1) * mask[i].view(
                batch_size, 1
            ) + alpha * (1 - mask[i]).view(batch_size, 1)

        # Every sequence needs to end with a transition to the stop_tag.
        if self.include_start_end_transitions:
            stops = alpha + self.end_transitions.view(1, num_tags)
        else:
            stops = alpha

        # Finally we log_sum_exp along the num_tags dim, result is (batch_size,)
        return torch.logsumexp(stops, -1)

    def _joint_likelihood(
        self, logits: torch.Tensor, tags: torch.Tensor, mask: torch.LongTensor
    ) -> torch.Tensor:
        """
        Computes the numerator term for the log-likelihood, which is just
        score(inputs, tags)
        """
        batch_size, sequence_length, _ = logits.data.shape

        # Transpose batch size and sequence dimensions:
        logits = logits.transpose(0, 1).contiguous()
        mask = mask.float().transpose(0, 1).contiguous()
        tags = tags.transpose(0, 1).contiguous()

        # Start with the transition scores from start_tag to the first tag in
        # each input
        score: Union[float, torch.Tensor]

        if self.include_start_end_transitions:
            score = self.start_transitions.index_select(0, tags[0])
        else:
            score = 0.0

        # Add up the scores for the observed transitions and all the inputs
        # but the last
        for i in range(sequence_length - 1):
            # Each is shape (batch_size,)
            current_tag, next_tag = tags[i], tags[i + 1]

            # The scores for transitioning from current_tag to next_tag
            transition_score = self.transitions[
                current_tag.view(-1), next_tag.view(-1)
            ]

            # The score for using current_tag
            emit_score = (
                logits[i].gather(1, current_tag.view(batch_size, 1)).squeeze(1)
            )

            # Include transition score if next element is unmasked,
            # input_score if this element is unmasked.
            score = (
                score + transition_score * mask[i + 1] + emit_score * mask[i]
            )

        # Transition from last state to "stop" state. To start with, we need
        # to find the last tag
        # for each instance.
        last_tag_index = mask.sum(0).long() - 1
        last_tags = tags.gather(0, last_tag_index.view(1, batch_size)).squeeze(
            0
        )

        # Compute score of transitioning to `stop_tag` from each "last tag".
        last_transition_score: Union[float, torch.Tensor]

        if self.include_start_end_transitions:
            last_transition_score = self.end_transitions.index_select(
                0, last_tags
            )
        else:
            last_transition_score = 0.0

        # Add the last input if it's not masked.
        last_inputs = logits[-1]  # (batch_size, num_tags)
        last_input_score = last_inputs.gather(
            1, last_tags.view(-1, 1)
        )  # (batch_size, 1)
        last_input_score = last_input_score.squeeze()  # (batch_size,)

        score = score + last_transition_score + last_input_score * mask[-1]

        return score

    def forward(
        self,
        inputs: torch.Tensor,
        tags: torch.Tensor,
        mask: Optional[torch.ByteTensor] = None,
    ) -> torch.Tensor:
        """
        Computes the log likelihood.
        """
        # pylint: disable=arguments-differ
        if mask is None:
            mask = torch.ones(*tags.size(), dtype=torch.long)

        log_denominator = self._input_likelihood(inputs, mask)
        log_numerator = self._joint_likelihood(inputs, tags, mask)

        return torch.sum(log_numerator - log_denominator)

    def viterbi_tags(
        self, logits: torch.Tensor, mask: torch.Tensor
    ) -> List[Tuple[List[int], float]]:
        """
        Uses viterbi algorithm to find most likely tags for the given inputs.
        If constraints are applied, disallows all other transitions.
        """
        _, max_seq_length, num_tags = logits.size()

        # Get the tensors out of the variables
        logits, mask = logits.data, mask.data

        # Augment transitions matrix with start and end transitions
        start_tag = num_tags
        end_tag = num_tags + 1
        transitions = torch.Tensor(num_tags + 2, num_tags + 2).fill_(-10000.0)

        # Apply transition constraints
        constrained_transitions = self.transitions * self._constraint_mask[
            :num_tags, :num_tags
        ] + -10000.0 * (1 - self._constraint_mask[:num_tags, :num_tags])
        transitions[:num_tags, :num_tags] = constrained_transitions.data

        if self.include_start_end_transitions:
            transitions[
                start_tag, :num_tags
            ] = self.start_transitions.detach() * self._constraint_mask[
                start_tag, :num_tags
            ].data + -10000.0 * (
                1 - self._constraint_mask[start_tag, :num_tags].detach()
            )
            transitions[
                :num_tags, end_tag
            ] = self.end_transitions.detach() * self._constraint_mask[
                :num_tags, end_tag
            ].data + -10000.0 * (
                1 - self._constraint_mask[:num_tags, end_tag].detach()
            )
        else:
            transitions[start_tag, :num_tags] = -10000.0 * (
                1 - self._constraint_mask[start_tag, :num_tags].detach()
            )
            transitions[:num_tags, end_tag] = -10000.0 * (
                1 - self._constraint_mask[:num_tags, end_tag].detach()
            )

        best_paths = []
        # Pad the max sequence length by 2 to account for start_tag + end_tag.
        tag_sequence = torch.Tensor(max_seq_length + 2, num_tags + 2)

        for prediction, prediction_mask in zip(logits, mask):
            sequence_length = torch.sum(prediction_mask)

            # Start with everything totally unlikely
            tag_sequence.fill_(-10000.0)
            # At timestep 0 we must have the START_TAG
            tag_sequence[0, start_tag] = 0.0
            # At steps 1, ..., sequence_length we just use the
            # incoming prediction
            tag_sequence[1 : (sequence_length + 1), :num_tags] = prediction[
                :sequence_length
            ]
            # And at the last timestep we must have the END_TAG
            tag_sequence[sequence_length + 1, end_tag] = 0.0

            # We pass the tags and the transitions to ``viterbi_decode``.
            viterbi_path, viterbi_score = viterbi_decode(
                tag_sequence[: (sequence_length + 2)], transitions
            )
            # Get rid of START and END sentinels and append.
            viterbi_path = viterbi_path[1:-1]
            best_paths.append((viterbi_path, viterbi_score.item()))

        return best_paths


def viterbi_decode(
    tag_sequence: torch.Tensor,
    transition_matrix: torch.Tensor,
    tag_observations: Optional[List[int]] = None,
):
    """
    Perform Viterbi decoding in log space over a sequence given a transition
    matrix specifying pairwise (transition) potentials between tags and a
    matrix of shape (sequence_length, num_tags) specifying unary potentials for
    possible tags per timestep.

    Parameters
    ----------
    tag_sequence : torch.Tensor, required.
        A tensor of shape (sequence_length, num_tags) representing scores for
        a set of tags over a given sequence.
    transition_matrix : torch.Tensor, required.
        A tensor of shape (num_tags, num_tags) representing the binary
        potentials for transitioning between a given pair of tags.
    tag_observations : Optional[List[int]], optional, (default = None)
        A list of length ``sequence_length`` containing the class ids of
        observed elements in the sequence, with unobserved elements being set
        to -1.
        Note that it is possible to provide evidence which results in
        degenerate labelings if the sequences of tags you provide as evidence
        cannot transition between each other, or those transitions are
        extremely unlikely. In this situation we log a warning, but the
        responsibility for providing self-consistent evidence
        ultimately lies with the user.

    Returns
    -------
    viterbi_path : List[int]
        The tag indices of the maximum likelihood tag sequence.
    viterbi_score : torch.Tensor
        The score of the viterbi path.
    """
    sequence_length, num_tags = list(tag_sequence.size())
    if tag_observations:
        if len(tag_observations) != sequence_length:
            raise ValueError(
                "Observations were provided, but they were not the same length "
                "as the sequence. "
                "Found sequence of length: {} and evidence: {}".format(
                    sequence_length, tag_observations
                )
            )
    else:
        tag_observations = [-1 for _ in range(sequence_length)]

    path_scores = []
    path_indices = []

    if tag_observations[0] != -1:
        one_hot = torch.zeros(num_tags)
        one_hot[tag_observations[0]] = 100000.0
        path_scores.append(one_hot)
    else:
        path_scores.append(tag_sequence[0, :])

    # Evaluate the scores for all possible paths.
    for timestep in range(1, sequence_length):
        # Add pairwise potentials to current scores.
        summed_potentials = (
            path_scores[timestep - 1].unsqueeze(-1) + transition_matrix
        )
        scores, paths = torch.max(summed_potentials, 0)

        # If we have an observation for this timestep, use it
        # instead of the distribution over tags.
        observation = tag_observations[timestep]
        # Warn the user if they have passed
        # invalid/extremely unlikely evidence.
        if tag_observations[timestep - 1] != -1:
            if (
                transition_matrix[tag_observations[timestep - 1], observation]
                < -10000
            ):
                logger.warning(
                    "The pairwise potential between tags you have passed as "
                    "observations is extremely unlikely. Double check your"
                    "evidence or transition potentials!"
                )
        if observation != -1:
            one_hot = torch.zeros(num_tags)
            one_hot[observation] = 100000.0
            path_scores.append(one_hot)
        else:
            path_scores.append(tag_sequence[timestep, :] + scores.squeeze())
        path_indices.append(paths.squeeze())

    # Construct the most likely sequence backwards.
    viterbi_score, best_path = torch.max(path_scores[-1], 0)
    viterbi_path = [int(best_path.numpy())]
    for backward_timestep in reversed(path_indices):
        viterbi_path.append(int(backward_timestep[viterbi_path[-1]]))
    # Reverse the backward path.
    viterbi_path.reverse()
    return viterbi_path, viterbi_score



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=non-parent-init-called
# pylint: disable=super-init-not-called
"""
A magic model that registers the parameter of a pytorch nn module
and performs memory-efficient parameter updates locally.
"""

import copy
from typing import Dict
from forte.utils import create_import_error_msg

try:
    import torch
    from torch import nn
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "magic models")
    ) from e

try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg("texar-pytorch", "models", "magic models")
    ) from e

__all__ = ["MetaModule", "TexarBertMetaModule"]


class MetaModule(nn.ModuleList):
    # pylint: disable=line-too-long
    r"""A class extending :class:`torch.nn.ModuleList`
    that registers the parameters of a :class:`torch.nn.Module`
    and performs memory-efficient parameter updates locally.

    This code is adapted from:
    https://github.com/tanyuqian/learning-data-manipulation/blob/master/magic_module.py

    It implements the calculation:
    :math:`L(\theta - \nabla_{\theta} L_{train}(\theta, \phi))`.

    Args:
        module: A :class:`torch.nn.Module`.

    This class can be used for simple input module, whose sub-modules don't
    contain other helper functions or attributes that do not belong to this
    class to perform their :meth:`forward`.

    Otherwise, since :meth:`forward` calls the input module's :meth:`forward`,
    in order to perform :meth:`forward` of the sub-modules of the input module
    correctly, this class needs to extend those sub-modules that define
    the methods needed for their :meth:`forward`, so that it inherits their
    methods to perform the sub-module's :meth:`forward`.

    For example, if the input module is
    :class:`~texar.torch.modules.BERTClassifier`,
    :meth:`_get_noise_shape`, :meth:`_split_heads`, :meth:`_combine_heads`
    from its sub-modules (E.g. :class:`~texar.torch.modules.BERTEncoder`)
    are needed to be
    exposed in this class to perform their :meth:`forward`. Please refer to
    :class:`TexarBertMetaModule` for instructions on creating a subclass from
    this one for a specific input module.
    """

    def __init__(self, module: nn.Module):
        nn.Module.__init__(self)
        self._type = type(module)

        for key, value in module._parameters.items():
            if value is not None:
                self.register_parameter("_origin_" + key, value)
                self.register_buffer(key, value.data)
            else:
                self.register_buffer(key, None)

        for key, value in module._buffers.items():
            self.register_buffer(key, copy.deepcopy(value))

        # Recursively create MetaModule.
        for key, value in module._modules.items():
            # type(self) is the real class object
            # it can be MetaModule(value), or it can be its subclass,
            # e.g. TexarBertMetaModule(value)
            self.add_module(key, type(self)(value))

        for key, value in module.__dict__.items():
            if (
                key not in self.__dict__
                and key not in self._buffers
                and key not in self._modules
            ):
                self.__setattr__(key, value)

    def forward(self, *args, **kwargs):
        return self._type.forward(self, *args, **kwargs)

    def update_params(self, deltas: Dict[str, torch.Tensor]):
        sub_params: Dict[str, torch.Tensor] = {}
        for key, delta in deltas.items():
            if "." not in key:
                self._buffers[key] = self._buffers[key] + delta
            else:
                attr = key.split(".")[0]
                if attr not in sub_params:
                    sub_params[attr] = {}
                sub_params[attr][".".join(key.split(".")[1:])] = delta
        for key, value in sub_params.items():
            self._modules[key].update_params(value)


class TexarBertMetaModule(
    MetaModule, tx.modules.EmbedderBase, tx.modules.MultiheadAttentionEncoder
):
    r"""A subclass that extends :class:`MetaModule` to do parameter updates
    locally for texar-pytorch Bert related modules.
    E.g. :class:`texar.torch.modules.BERTClassifier`

    Please refer to its base class :class:`MetaModule` for more details.

    Args:
        module: A :class:`torch.nn.Module`.

    This class extends :class:`~texar.torch.modules.EmbedderBase` and
    :class:`~texar.torch.modules.MultiheadAttentionEncoder`, such that it
    inherits their methods that are needed to perform :meth:`forward` of
    the modules that utilizes these methods,
    E.g. :class:`~texar.torch.modules.BERTEncoder`,

    Some notes of the order of the base classes that this class extends:

    `MetaModule` should be the first one, so that its :meth:`forward` will
    call :meth:`MetaModule.forward` instead of the :meth:`forward` of the other
    base classes, such as
    :func:`texar.torch.modules.MultiheadAttentionEncoder.forward`.
    If `MetaModule` is not the first one, then a :meth:`forward` should be
    defined in this class, such that it is called correctly.

    Example:

        .. code-block:: python

            def forward(self, *args, **kwargs):
                return MetaModule.forward(self, *args, **kwargs)

    """

    def __init__(self, module: nn.Module):
        MetaModule.__init__(self, module)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads Ontonotes data into Datapacks.
"""
import os
from collections import defaultdict
from typing import (
    Any,
    DefaultDict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
)

from forte.common.exception import ProcessorConfigError
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import (
    CoreferenceGroup,
    Document,
    EntityMention,
    PredicateArgument,
    PredicateLink,
    PredicateMention,
    Sentence,
    Token,
)

__all__ = [
    "OntonotesReader",
]

Stack = List[Tuple[int, str]]


class OntonotesReader(PackReader):
    r""":class:`~forte.data.readers.ontonotes_reader.OntonotesReader` is
    designed to read in the English OntoNotes
    v5.0 data in the datasets used by the CoNLL 2011/2012 shared tasks. To use
    this Reader, you must follow the instructions provided `here (v12 release):
    <http://cemantix.org/data/ontonotes.html>`_:, which will allow you to
    download the CoNLL style annotations for the OntoNotes v5.0 release
    – LDC2013T19.tgz obtained from LDC.

    """

    # NOTE: sphinx 4.0 doesn't like inner class, making it protected to skip.
    class _ParsedFields(NamedTuple):
        """
        Internal class used to specify the fields of an `ontonotes` file.
        """

        word: str
        predicate_labels: List[str] = []
        document_id: Optional[str] = None
        part_number: Optional[str] = None
        pos_tag: Optional[str] = None
        lemmatised_word: Optional[str] = None
        framenet_id: Optional[str] = None
        word_sense: Optional[str] = None
        speaker: Optional[str] = None
        entity_label: Optional[str] = None
        coreference: Optional[str] = None

    _DEFAULT_FORMAT = [
        "document_id",
        "part_number",
        None,
        "word",
        "pos_tag",
        None,
        "lemmatised_word",
        "framenet_id",
        "word_sense",
        "speaker",
        "entity_label",
        "*predicate_labels",
        "coreference",
    ]
    _STAR_FIELDS = {"predicate_labels"}
    _REQUIRED_FIELDS = ["word", "predicate_labels"]

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if configs.column_format is None:
            raise ProcessorConfigError(
                "Configuration column_format not provided."
            )

        column_format = configs.column_format
        # Validate column format.
        seen_fields: Set[str] = set()
        # pylint: disable=attribute-defined-outside-init
        self._column_format: List[Optional[str]] = []
        self._star_pos = None
        for idx, field in enumerate(column_format):
            if field is None:
                self._column_format.append(None)
                continue
            if field.startswith("*"):
                if self._star_pos is not None:
                    raise ValueError("Only one field can begin with '*'")
                field = field[1:]
                if field not in self._STAR_FIELDS:
                    raise ValueError(f"Field '{field}' cannot begin with '*'")
                self._star_pos = idx
            if field not in self._ParsedFields._fields:
                raise ValueError(f"Unsupported field type: '{field}'")
            if field in seen_fields:
                raise ValueError(f"Duplicate field type: '{field}'")
            seen_fields.add(field)
            self._column_format.append(field)
        # Sanity check: certain fields must be present in format.
        for field in self._REQUIRED_FIELDS:
            if field not in seen_fields:
                raise ValueError(f"'{field}' field is required")

    @classmethod
    def default_configs(cls):
        r"""
        Returns a dictionary of default hyperparameters.

        .. code-block:: python

            {
                "name": "reader",
                "column_format": [
                    "document_id",
                    "part_number",
                    None,
                    "word",
                    "pos_tag",
                    None,
                    "lemmatised_word",
                    "framenet_id",
                    "word_sense",
                    "speaker",
                    "entity_label",
                    "*predicate_labels",
                    "coreference",
                ]
            }

        Here:

        - `"column_format"`: list
          A list of strings indicating which field each column in a
          line corresponds to. The length of the list should be equal to the
          number of columns in the files to be read. Available field types
          include:

            - ``"document_id"``
            - ``"part_number"``
            - ``"word"``
            - ``"pos_tag"``
            - ``"lemmatised_word"``
            - ``"framenet_id"``
            - ``"word_sense"``
            - ``"speaker"``
            - ``"entity_label"``
            - ``"coreference"``
            - ``"*predicate_labels"``

            Field types marked with ``*`` indicate a variable-column field: it
            could span multiple fields. Only one such field is allowed in the
            format specification.

            If a column should be ignored, fill in `None` at the corresponding
            position.

            .. note::
                A `None` field means that column in the dataset file will be
                ignored during parsing.
        """
        config: dict = super().default_configs()

        config.update({"column_format": cls._DEFAULT_FORMAT})
        return config

    def _collect(self, conll_directory: str) -> Iterator[Any]:  # type: ignore
        r"""Iterator over `*.gold_conll` files in the data_source

        Args:
            conll_directory:  path to the directory containing the files.

        Returns: Iterator over files with gold_conll path.
        """
        return dataset_path_iterator(conll_directory, "gold_conll")

    def _cache_key_function(self, conll_file: str) -> str:
        return os.path.basename(conll_file)

    def _parse_line(self, line: str) -> "_ParsedFields":
        parts = line.split()
        fields = {}
        if self._star_pos is not None:
            l = self._star_pos
            r = len(parts) - (len(self._column_format) - self._star_pos - 1)
            parts = parts[:l] + [parts[l:r]] + parts[r:]  # type: ignore
        for field, part in zip(self._column_format, parts):
            if field is not None:
                fields[field] = part
        return self._ParsedFields(**fields)  # type: ignore

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        start_new_doc: bool = True

        with open(file_path, encoding="utf8") as doc:
            for line in doc:
                if start_new_doc:
                    pack = DataPack()

                    words: List = []
                    offset = 0
                    has_rows = False

                    speaker = part_id = document_id = None
                    sentence_begin = 0

                    # auxiliary structures
                    current_entity_mention: Optional[Tuple[int, str]] = None
                    verbal_predicates: List[PredicateMention] = []

                    current_pred_arg: List[Stack] = []
                    verbal_pred_args: List[
                        List[Tuple[PredicateArgument, str]]
                    ] = []

                    groups: DefaultDict[int, List[EntityMention]] = defaultdict(
                        list
                    )
                    coref_stacks: DefaultDict[int, List[int]] = defaultdict(
                        list
                    )

                    start_new_doc = False

                line = line.strip()

                if line.startswith("#end document"):
                    # Group the coreference mentions in the whole document.
                    for _, mention_list in groups.items():
                        group = CoreferenceGroup(pack)
                        group.add_members(mention_list)

                    text = " ".join(words)
                    pack.set_text(
                        text, replace_func=self.text_replace_operation
                    )

                    _ = Document(pack, 0, len(text))
                    if document_id is not None:
                        pack.pack_name = document_id

                    yield pack

                    start_new_doc = True
                    continue

                if line != "" and not line.startswith("#"):
                    fields = self._parse_line(line)
                    speaker = fields.speaker
                    if fields.part_number is not None:
                        part_id = int(fields.part_number)
                    document_id = fields.document_id

                    assert fields.word is not None
                    word_begin = offset
                    word_end = offset + len(fields.word)

                    # add tokens
                    token = Token(pack, word_begin, word_end)

                    if fields.pos_tag is not None:
                        token.pos = fields.pos_tag
                    if fields.word_sense is not None:
                        token.sense = fields.word_sense

                    # add entity mentions
                    current_entity_mention = self._process_entity_annotations(
                        pack,
                        fields.entity_label,
                        word_begin,
                        word_end,
                        current_entity_mention,
                    )

                    # add predicate mentions
                    if (
                        fields.lemmatised_word is not None
                        and fields.lemmatised_word != "-"
                    ):
                        word_is_verbal_predicate = any(
                            "(V" in x for x in fields.predicate_labels
                        )
                        pred_mention = PredicateMention(
                            pack, word_begin, word_end
                        )

                        pred_mention.predicate_lemma = fields.lemmatised_word
                        pred_mention.is_verb = word_is_verbal_predicate

                        if fields.framenet_id is not None:
                            pred_mention.framenet_id = fields.framenet_id

                        if word_is_verbal_predicate:
                            verbal_predicates.append(pred_mention)

                    if not verbal_pred_args:
                        current_pred_arg = [[] for _ in fields.predicate_labels]
                        verbal_pred_args = [[] for _ in fields.predicate_labels]

                    # add predicate arguments
                    self._process_pred_annotations(
                        pack,
                        fields.predicate_labels,
                        word_begin,
                        word_end,
                        current_pred_arg,
                        verbal_pred_args,
                    )

                    # add coreference mentions
                    self._process_coref_annotations(
                        pack,
                        fields.coreference,
                        word_begin,
                        word_end,
                        coref_stacks,
                        groups,
                    )

                    words.append(fields.word)
                    offset = word_end + 1
                    has_rows = True

                else:
                    if not has_rows:
                        continue

                    # add predicate links in the sentence
                    for predicate, pred_arg in zip(
                        verbal_predicates, verbal_pred_args
                    ):
                        for arg in pred_arg:
                            link = PredicateLink(pack, predicate, arg[0])
                            link.arg_type = arg[1]

                    verbal_predicates = []
                    current_pred_arg = []
                    verbal_pred_args = []

                    # add sentence

                    sent = Sentence(pack, sentence_begin, offset - 1)
                    if speaker is not None:
                        sent.speaker = speaker
                    if part_id is not None:
                        sent.part_id = int(part_id)

                    sentence_begin = offset

                    has_rows = False

    def _process_entity_annotations(
        self,
        pack: DataPack,
        label: Optional[str],
        word_begin: int,
        word_end: int,
        current_entity_mention: Optional[Tuple[int, str]],
    ) -> Optional[Tuple[int, str]]:
        if label is None:
            return None

        ner_type = label.strip("()*")

        if "(" in label:
            # Entering into a span for a particular ner.
            current_entity_mention = (word_begin, ner_type)
        if ")" in label:
            if current_entity_mention is None:
                raise ValueError(
                    "current_entity_mention is None when meet right blanket."
                )
            # Exiting a span, add and then reset the current span.
            entity = EntityMention(pack, current_entity_mention[0], word_end)
            entity.ner_type = current_entity_mention[1]

            current_entity_mention = None

        return current_entity_mention

    def _process_pred_annotations(
        self,
        pack: DataPack,
        labels: List[str],
        word_begin: int,
        word_end: int,
        current_pred_arg: List[Stack],
        verbal_pred_args: List[List[Tuple[PredicateArgument, str]]],
    ) -> None:
        """
        Various cases will be handled regarding nested spans including:
        case 1:
        (xxx(xxx*)*)

        case 2:
        (xxx(xxx*)
        *
        *)

        case 3:
        (xxx(xxx*
        *
        *)
        *
        *)

        case 4:
        (xxx*
        *
        (xxx
        *
        *)
        *
        *)
        Where the `xxx` represents an argument type.
        A stack will be maintained to parse nested spans.
        """
        # TODO: currently all nested spans will be parsed but only outer most
        #  span will be stored into datapack.
        for label_index, label in enumerate(labels):
            label = label.strip()

            arg_type: str = ""
            i: int = 0
            while i < len(label):
                c: str = label[i]
                if c == "*":
                    i += 1
                elif c == "(":
                    # New argument span.
                    j: int = i + 1
                    while j < len(label):
                        arg_type += label[j]
                        if j == len(label) - 1 or label[j + 1] in ["(", ")"]:
                            break
                        j += 1
                    i = j + 1
                    arg_type = arg_type.strip("* ")

                    stack: Stack = current_pred_arg[label_index]
                    stack.append((word_begin, arg_type))
                    arg_type = ""
                elif c == ")":
                    # End of current argument span.
                    stack = current_pred_arg[label_index]
                    assert (
                        len(stack) > 0
                    ), "invalid parsing state: mismatch argument span"

                    arg_begin, arg_type = stack.pop()

                    if not stack:
                        # The outer most span will be stored into data pack
                        if arg_type != "V":
                            pred_arg = PredicateArgument(
                                pack, arg_begin, word_end
                            )

                            verbal_pred_args[label_index].append(
                                (pred_arg, arg_type)
                            )
                    i += 1

    def _process_coref_annotations(
        self,
        pack: DataPack,
        label: Optional[str],
        word_begin: int,
        word_end: int,
        coref_stacks: DefaultDict[int, List[int]],
        groups: DefaultDict[int, List[EntityMention]],
    ) -> None:

        if label is None or label == "-":
            return
        for segment in label.split("|"):
            # The conll representation of coref spans allows spans to overlap.
            if segment[0] == "(":
                if segment[-1] == ")":
                    # The span begins and ends at this word (single word span).
                    group_id = int(segment[1:-1])

                    coref_mention = EntityMention(pack, word_begin, word_end)
                    groups[group_id].append(coref_mention)
                else:
                    # The span is starting, so we record the index of the word.
                    group_id = int(segment[1:])
                    coref_stacks[group_id].append(word_begin)
            else:
                # The span for this id is ending, but not start at this word.
                group_id = int(segment[:-1])
                start = coref_stacks[group_id].pop()
                coref_mention = EntityMention(pack, start, word_end)
                groups[group_id].append(coref_mention)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads Stanford Sentiment Treebank
https://nlp.stanford.edu/sentiment/treebank.html
into data_pack format
"""
import os
import errno

from typing import Iterator, Dict, List, Tuple

from ft.onto.base_ontology import Sentence, ConstituentNode

from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader

__all__ = ["SST2Reader"]


class SST2Reader(PackReader):
    r""":class:`SST2Reader` is designed to read in the
    Stanford Sentiment Treebank 2 dataset from
    https://nlp.stanford.edu/sentiment/treebank.html
    To use the dataset, please download it from the webpage.

    Provided the ss2_dir_path, the _collect function will look for files:
        "dictionary.txt": a mapping from phrase text to phrase id
        "sentiment_labels": a mapping from phrase id to sentiment score
        "datasetSentences.txt": original text for the sentence
        "STree.txt": parent pointer list of constituency tree for each sentence

    """

    def __init__(self):
        super().__init__()
        # Transform the text-form phrase to sentiment score.
        self.phrase_to_id: Dict = {}
        self.id_to_senti: Dict = {}

    def _cache_key_function(self, data_pack: DataPack) -> str:
        if data_pack.pack_name is None:
            raise ValueError("data_pack does not have a sentence id")
        return data_pack.pack_name

    def _check_file_exist(self, filename: str):
        if not os.path.exists(filename):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), filename
            )

    def _collect(
        self, *args, **kwargs
    ) -> Iterator[List[Tuple[str, str, List[int]]]]:
        # pylint: disable = unused-argument
        r"""Iterator over sst files in the data_source.
        The directory should at least have the following files:
            "dictionary.txt": a mapping from phrase text to phrase id
            "sentiment_labels": a mapping from phrase id to sentiment score
            "datasetSentences.txt": original text for the sentence
            "STree.txt": parent pointer list of constituency tree
        Args:
            args: args[0] is the directory to the sst2 files.
                args[1] is the number of sentences in a data pack
            kwargs:
        Returns: data packs obtained from each sentence from the sst2 file.
        """
        sent_id: str
        sent_text: str
        parent_pointer_list: List[int]

        sst2_dir_path: str = args[0]
        n_samples: int = args[1]
        phrase_to_id_path: str = os.path.join(sst2_dir_path, "dictionary.txt")
        id_to_senti_path: str = os.path.join(
            sst2_dir_path, "sentiment_labels.txt"
        )
        text_path: str = os.path.join(sst2_dir_path, "datasetSentences.txt")
        tree_path: str = os.path.join(sst2_dir_path, "STree.txt")
        self._check_file_exist(phrase_to_id_path)
        self._check_file_exist(id_to_senti_path)
        self._check_file_exist(text_path)
        self._check_file_exist(tree_path)

        # Read the mapping from phrase to phrase-id.
        with open(phrase_to_id_path, "r", encoding="utf8") as file:
            for line in file:
                phrase, id_ = line.split("|")
                self.phrase_to_id[phrase] = int(id_)

        # Read the mapping from phrase-id to sentiment score.
        with open(id_to_senti_path, "r", encoding="utf8") as file:
            for i, line in enumerate(file):
                if i == 0:
                    continue
                id_, score = line.split("|")
                self.id_to_senti[int(id_)] = float(score)

        sent_lines = []
        # Read the text and tree structure.
        with open(text_path, "r", encoding="utf8") as ftext, open(
            tree_path, "r", encoding="utf8"
        ) as ftree:
            ftext.readline()  # Skip the headers.
            for line_text, line_tree in zip(ftext, ftree):
                line_text = line_text.strip()
                sent_id, sent_text = line_text.split("\t")
                parent_pointer_list = list(map(int, line_tree.split("|")))
                sent_lines.append((sent_id, sent_text, parent_pointer_list))
                # Yield a batch of sentences.
                if len(sent_lines) == n_samples:
                    yield sent_lines
                    sent_lines = []
            if len(sent_lines) > 0:
                yield sent_lines

    def _get_span_with_dfs(
        self,
        span_begin_end: List[List[int]],
        children_nodes: List[List[int]],
        cur_node: int,
    ):
        r"""Recursively get the span for each node in the tree
        Args:
            span_begin_end: stores the span (begin, end) posititon
            children_nodes: the structure of the tree
            cur_node: current processing node
        Returns: None
        """
        if len(children_nodes[cur_node]) == 0:
            return
        begin = -1
        end = -1
        for child in children_nodes[cur_node]:
            self._get_span_with_dfs(span_begin_end, children_nodes, child)
            if begin == -1 or begin > span_begin_end[child][0]:
                begin = span_begin_end[child][0]
            if end == -1 or end < span_begin_end[child][1]:
                end = span_begin_end[child][1]
        span_begin_end[cur_node] = [begin, end]

    def _parse_parent_pointer_list(
        self,
        data_pack: DataPack,
        sent_bias: int,
        sent_text: str,
        parent_pointer_list: List[int],
    ):
        r"""Build the ConstituentNode objects from parent pointer list.
        Args:
            data_pack: the data_pack to add ConstituentNode
            sent_bias: the start position of the sentence in the pack
            sent_text: the whitespace-splitted sentence text
            parent_pointer_list: a format to store the constituency tree
        Returns: None
        """

        tokens: List[str] = sent_text.split()
        n_nodes: int = len(parent_pointer_list) + 1
        span_begin_end: List[List[int]] = [[] for _ in range(n_nodes)]
        node_list: List[ConstituentNode]

        # Get the children node ids for each node, node 0 is the root.
        children_nodes: List[List[int]] = [[] for _ in range(n_nodes)]
        for i in range(1, n_nodes):
            parent = parent_pointer_list[i - 1]
            children_nodes[parent].append(i)

        # Get the begin/end index of spans for leaf nodes.
        for i in range(1, len(tokens) + 1):
            span_begin = 0
            if i > 1:
                # Plus 1 for the whitespace separator
                span_begin = span_begin_end[i - 1][1] + 1
            span_end = span_begin + len(tokens[i - 1])
            span_begin_end[i] = [span_begin, span_end]

        # Get the spans recursively and store in "span_begin_end".
        self._get_span_with_dfs(span_begin_end, children_nodes, 0)

        # Create the constituency Tree.
        node_list = [
            ConstituentNode(data_pack, begin + sent_bias, end + sent_bias)
            for (begin, end) in span_begin_end
        ]
        # Get the sentiment scores.
        for i in range(n_nodes):
            phrase = node_list[i].text
            phrase_id = self.phrase_to_id.get(phrase, -1)
            node_list[i].sentiment["pos"] = self.id_to_senti.get(phrase_id, 0.5)

        # Link the parent and children nodes.
        for i in range(1, n_nodes):
            parent = parent_pointer_list[i - 1]
            node_list[i].parent_node = node_list[parent]

        for i in range(n_nodes):
            # Sort the children nodes by span begin position.
            children_nodes[i].sort(key=lambda x: node_list[x].begin)
            for child in children_nodes[i]:
                node_list[i].children_nodes.append(node_list[child])

        # Set the is_leaf/is_root flag.
        for i in range(n_nodes):
            node_list[i].is_leaf = False
            node_list[i].is_root = False

        for i in range(1, len(tokens) + 1):
            node_list[i].is_leaf = True
        node_list[0].is_root = True

    def _parse_pack(self, sent_lines) -> Iterator[DataPack]:
        data_pack: DataPack = DataPack()
        sent_bias: int = 0
        batch_text: str = "\n".join(
            [sent_text for _, sent_text, _ in sent_lines]
        )
        data_pack.set_text(batch_text)

        for i, sent_line in enumerate(sent_lines):
            sent_id: str = sent_line[0]
            sent_text: str = sent_line[1].strip()
            parent_pointer_list: List[int] = sent_line[2]
            # Name the data_pack with the first sentence id.
            if i == 0:
                data_pack.pack_name = sent_id
            # Add sentence to data_pack.
            Sentence(data_pack, sent_bias, sent_bias + len(sent_text))
            self._parse_parent_pointer_list(
                data_pack, sent_bias, sent_text, parent_pointer_list
            )

            sent_bias += len(sent_text) + 1

        yield data_pack



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader for reading sentences from text files into MultiPack
"""
import os
from typing import Any, Iterator, Dict, Tuple

from forte.data.data_utils_io import dataset_path_iterator_with_base
from forte.data.multi_pack import MultiPack
from forte.data.base_reader import MultiPackReader
from ft.onto.base_ontology import Sentence

__all__ = ["MultiPackSentenceReader"]


class MultiPackSentenceReader(MultiPackReader):
    r""":class:`MultiPackSentenceReader` is designed to read a directory of
    files and convert each file's contents into a data pack. This class yields a
    multipack with pack ``input_pack_name`` containing the file's contents.
    It additionally packs an empty pack with name ``output_pack_name`` into the
    multipack.
    """

    def _collect(self, text_directory: str) -> Iterator[Any]:  # type: ignore
        return dataset_path_iterator_with_base(text_directory, "")

    def _cache_key_function(self, txt_path: str) -> str:
        return os.path.basename(txt_path)

    def _parse_pack(
        self, base_and_path: Tuple[str, str]
    ) -> Iterator[MultiPack]:
        base_dir, file_path = base_and_path

        m_pack: MultiPack = MultiPack()

        input_pack_name = self.configs.input_pack_name
        output_pack_name = self.configs.output_pack_name

        text = ""
        offset = 0
        with open(file_path, "r", encoding="utf8") as doc:
            # Remove long path from the beginning.
            doc_id = file_path[
                file_path.startswith(base_dir) and len(base_dir) :
            ]
            doc_id = doc_id.strip(os.path.sep)

            input_pack = m_pack.add_pack(input_pack_name)
            input_pack.pack_name = doc_id

            for line in doc:
                line = line.strip()

                if len(line) == 0:
                    continue

                # add sentence
                Sentence(input_pack, offset, offset + len(line))
                text += line + "\n"
                offset = offset + len(line) + 1

            input_pack.set_text(text, replace_func=self.text_replace_operation)
            # Create a output pack without text.
            m_pack.add_pack(output_pack_name)
            yield m_pack

    @classmethod
    def default_configs(cls) -> Dict[str, str]:
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                "name": "multipack_sentence_reader"
                "input_pack_name": "input_src",
                "output_pack_name": "output_tgt"
            }

        Here:

            - `"name"`: str
              Name of the reader

            - `"input_pack_name"`: str
              Name of the input pack. This name can be used to retrieve the
              input pack from the multipack.

            - `"output_pack_name"`: str
              Name of the output pack. This name can be used to retrieve the
              output pack from the multipack.
        """
        return {
            "name": "multipack_sentence_reader",
            "input_pack_name": "input_src",
            "output_pack_name": "output_tgt",
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import defaultdict
from typing import (
    DefaultDict,
    Dict,
    List,
    Set,
    Type,
    Hashable,
    Generic,
    Iterable,
    Tuple,
    KeysView,
)

from forte.common.exception import PackIndexError
from forte.data.ontology.core import GroupType, LinkType, EntryType

logger = logging.getLogger(__name__)

__all__ = ["BaseIndex"]


class BaseIndex(Generic[EntryType]):
    r"""A set of indexes used in :class:`~forte.data.base_pack.BasePack`:

    #. :attr:`entry_index`, the index from each ``tid`` to the corresponding entry
    #. :attr:`type_index`, the index from each type to the entries of
       that type
    #. :attr:`link_index`, the index from child
       (:attr:`link_index["child_index"]`)and parent
       (:attr:`link_index["parent_index"]`) nodes to links
    #. :attr:`group_index`, the index from group members to groups.

    """

    def __init__(self):
        # List of basic indexes (always on).

        # Mapping from entry's tid to the entries.
        self._entry_index: Dict[int, EntryType] = {}

        # Mapping from entry's type to entries' id.
        self._type_index: DefaultDict[Type[EntryType], Set[int]] = defaultdict(
            set
        )

        # Indices below will be built when looked up:

        # A cache map to store all entries of a certain type including the
        # sub-types. This index will be populated on demand when
        # query_by_type_subtype is called.
        self._subtype_index: Dict[Type[EntryType], Set[int]] = {}

        self._group_index: DefaultDict[Hashable, Set[int]] = defaultdict(set)
        self._link_index: Dict[str, DefaultDict[Hashable, Set[int]]] = {}

        # Indexing switches.
        self._group_index_switch = False
        self._link_index_switch = False

    def update_basic_index(self, entries: List[EntryType]):
        r"""Build or update the basic indexes, including

        (1) :attr:`entry_index`,
        the index from each ``tid`` to the corresponding entry;

        (2) :attr:`type_index`, the index from each type to the entries of that
        type;

        (3) :attr:`component_index`, the index from each component to the
        entries generated by that component.

        Args:
            entries (list): a list of entries to be added into the basic index.
        """
        for entry in entries:
            self._entry_index[entry.tid] = entry
            self._type_index[type(entry)].add(entry.tid)
            # Disable sub type index since new items are added and this will
            #  be rebuilt in next query (`query_by_type_subtype`).
            self._subtype_index.pop(type(entry), None)

    def get_entry(self, tid: int) -> EntryType:
        return self._entry_index[tid]

    def indexed_types(self) -> KeysView[Type]:
        return self._type_index.keys()

    def query_by_type(self, t: Type[EntryType]) -> Set[int]:
        return self._type_index[t]

    def query_by_type_subtype(self, t: Type[EntryType]) -> Set[int]:
        r"""Look up the entry indices that are instances of ``entry_type``,
        including children classes of ``entry_type``.

        .. note::

            all the known types to this data pack will be scanned to find
            all sub-types. This method will try to cache the sub-type information
            after the first call, but the cached information could be invalidated
            by other operations (such as adding new items to the data pack).

        Args:
            t: The type of the entry you are looking for.

        Returns:
             A set of entry ids. The entries are instances of ``entry_type`` (
             and also includes instances of the subclasses of `entry_type`).
        """
        if t in self._subtype_index:
            return self._subtype_index[t]
        else:
            subclass_index: Set[int] = set()
            for index_key, index_val in self.iter_type_index():
                if issubclass(index_key, t):
                    subclass_index.update(index_val)
            self._subtype_index[t] = subclass_index
            return subclass_index

    def iter_type_index(self) -> Iterable[Tuple[Type, Set[int]]]:
        for t, ids in self._type_index.items():
            yield t, ids

    def remove_entry(self, entry: EntryType):
        self._entry_index.pop(entry.tid)
        self._type_index[type(entry)].remove(entry.tid)

        self.turn_group_index_switch(on=False)
        self.turn_link_index_switch(on=False)

    @property
    def link_index_on(self):
        return self._link_index_switch

    def turn_link_index_switch(self, on: bool):
        self._link_index_switch = on

    @property
    def group_index_on(self):
        return self._group_index_switch

    def turn_group_index_switch(self, on: bool):
        self._group_index_switch = on

    def build_link_index(self, links: List[LinkType]):
        r"""Build the :attr:`link_index`, the index from child and parent nodes
        to links. It will build the links with the links in the dataset.

        :attr:`link_index` consists of two sub-indexes:
        "child_index" is the index from child nodes to their corresponding
        links, and "parent_index" is the index from parent nodes to their
        corresponding links.
        Returns:

        """
        self.turn_link_index_switch(on=True)
        self._link_index["child_index"] = defaultdict(set)
        self._link_index["parent_index"] = defaultdict(set)
        self.update_link_index(links)

    def build_group_index(self, groups: List[GroupType]):
        r"""Build :attr:`group_index`, the index from group members to groups.

        Returns:
            None
        """
        self.turn_group_index_switch(on=True)
        self._group_index = defaultdict(set)
        self.update_group_index(groups)

    def link_index(self, tid: int, as_parent: bool = True) -> Set[int]:
        r"""Look up the link_index with key ``tid``. If the link index is not
        built, this will throw a :class:`~forte.common.exception.PackIndexError`.

        Args:
            tid: the ``tid`` of the entry being looked up.
            as_parent: If `as_patent` is True, will look up
                :attr:`link_index["parent_index"] and return the tids of links
                whose parent is ``tid``. Otherwise,  will look up
                :attr:`link_index["child_index"] and return the tids of links
                whose child is ``tid``.
        """
        if not self._link_index_switch:
            raise PackIndexError("Link index for pack not build")

        if as_parent:
            return self._link_index["parent_index"][tid]
        else:
            return self._link_index["child_index"][tid]

    def group_index(self, tid: int) -> Set[int]:
        r"""Look up the group_index with key ``tid``. If the index is not built,
        this will raise a :class:`~forte.common.exception.PackIndexError`.
        """
        if not self.group_index_on:
            raise PackIndexError("Group index for pack not build")
        return self._group_index[tid]

    def update_link_index(self, links: List[LinkType]):
        r"""Update :attr:`link_index` with the provided links, the index from
        child and parent to links.

        :attr:`link_index` consists of two sub-indexes:

            - "child_index" is the index from child nodes to their corresponding
              links
            - "parent_index" is the index from parent nodes to their
              corresponding links.

        Args:
            links: a list of links to be added into the index.
        """
        if not self.link_index_on:
            raise PackIndexError("Link index has not been built.")

        for link in links:
            self._link_index["child_index"][link.get_child().index_key].add(
                link.index_key
            )
            self._link_index["parent_index"][link.get_parent().index_key].add(
                link.index_key
            )

    def update_group_index(self, groups: List[GroupType]):
        r"""Build or update :attr:`group_index`, the index from group members
        to groups.

        Args:
            groups: a list of groups to be added into the index.
        """
        logger.debug("Updating group index")

        if not self._group_index:
            raise PackIndexError("Group index has not been built.")

        for group in groups:
            for member in group.members:  # type: ignore
                self._group_index[member].add(group.tid)

    def add_link_parent(self, parent: EntryType, link: LinkType):
        self._link_index["parent_index"][parent.index_key].add(link.tid)

    def add_link_child(self, child: EntryType, link: LinkType):
        self._link_index["child_index"][child.index_key].add(link.tid)

    def add_group_member(self, member: EntryType, group: GroupType):
        self._group_index[member.index_key].add(group.tid)



#  Copyright 2020 The Forte Authors. All Rights Reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provide a framework inherited by users to do the training.
"""
import logging
from abc import abstractmethod
from typing import Dict, Iterator, Any, Optional
import pickle

from forte.data.data_pack import DataPack
from forte.train_preprocessor import TrainPreprocessor

logger = logging.getLogger(__name__)

__all__ = ["BaseTrainer"]


class BaseTrainer:
    r"""
    `BaseTrainer` is the main entry for using Forte training framework. Users
    should inherit this class and overwrite multiple methods defined in this
    class. Internally, it will make use of
    :class:`~forte.train_preprocessor.TrainPreprocessor` to do the the actual
    training. Please refer to the documentation of that class for details. A
    concrete example is provided below showing how to use this class.

    Below is an example for how to use this class. A fully documented example is
    also provided in
    :class:`~forte.examples.tagging.tagging_trainer.TaggingTrainer`.

    .. code-block:: python

        class TaggingTrainer(BaseTrainer):
            def create_tp_config(self) -> Dict:
                tp_request: Dict = {
                    "scope": Sentence,
                    "schemes": {
                        "text_tag": {
                            "type": TrainPreprocessor.DATA_INPUT,
                            "extractor": {
                                "class_name":
                                "forte.data.extractors.AttributeExtractor",
                                "config": {
                                    "entry_type": "ft.onto.base_ontology.Token",
                                    "vocab_method": "indexing",
                                    "attribute": "text"
                                }
                            }
                        },
                        "char_tag": {
                            "type": TrainPreprocessor.DATA_INPUT,
                            "extractor": {
                                "class_name":
                                "forte.data.extractors.CharExtractor",
                                "config": {
                                    "entry_type": "ft.onto.base_ontology.Token",
                                    "vocab_method": "indexing",
                                    "max_char_length": 45
                                }
                            }
                        },
                        "output_tag": {
                            "type": TrainPreprocessor.DATA_OUTPUT,
                            "extractor": {
                                "class_name":
                                "forte.data.extractors.BioSeqTaggingExtractor",
                                "config": {
                                    "vocab_method": "indexing"
                                    "pad_value": 0
                                }
                            }
                        }
                    }
                }
                return tp_request

            def create_pack_iterator(self) -> Iterator[DataPack]:
                reader = CoNLL03Reader()
                train_pl: Pipeline = Pipeline()
                train_pl.set_reader(reader)
                train_pl.initialize()
                pack_iterator: Iterator[DataPack] = \
                    train_pl.process_dataset(self.config_data.train_path)
                return pack_iterator

            def train(self):
                schemes: Dict = self.train_preprocessor.request["schemes"]
                text_extractor: BaseExtractor = \
                    schemes["text_tag"]["extractor"]
                char_extractor: BaseExtractor = \
                    schemes["char_tag"]["extractor"]
                output_extractor: BaseExtractor = \
                    schemes["output_tag"]["extractor"]

                model: BiRecurrentConvCRF = \
                    BiRecurrentConvCRF(word_vocab=text_extractor.get_dict(),
                                       char_vocab_size=char_extractor.size(),
                                       tag_vocab_size=output_extractor.size(),
                                       config_model=self.config_model)
                model.to(self.device)

                optim: Optimizer = SGD(model.parameters())

                tp = self.train_preprocessor

                # Train for 10 epochs
                for epoch in range(10):
                    # Get iterator of preprocessed batch of train data
                    batch_iter: Iterator[Batch] = tp.get_train_batch_iterator()

                    for batch in tqdm(batch_iter):
                        word = batch["text_tag"]["data"]
                        char = batch["char_tag"]["data"]
                        output = batch["output_tag"]["data"]
                        word_masks = batch["text_tag"]["masks"][0]

                        optim.zero_grad()

                        loss = model(word, char, output, mask=word_masks)

                        loss.backward()
                        optim.step()
    """

    def __init__(self):
        self._pack_iterator: Optional[Iterator[DataPack]] = None
        self._tp: Optional[TrainPreprocessor] = None
        self._initialized: bool = False

    def initialize(self):
        # Check if initialize has already been called before
        if self._initialized:
            return

        logging.info("Initializing the trainer...")
        self._pack_iterator: Iterator[DataPack] = self.create_pack_iterator()
        self._tp = TrainPreprocessor(pack_iterator=self._pack_iterator)
        self._tp.initialize(config=self.create_tp_config())
        self._initialized = True
        logging.info("Done Initializing.")

    @property
    def train_preprocessor(self) -> Optional[TrainPreprocessor]:
        r"""The instance of type
        :class:`~forte.train_preprocessor.TrainPreprocessor`. The Trainer will
        internally create an instance of this class to do the actual training.
        """
        if not self._initialized:
            raise ValueError(
                "initialize should be called to " "build train preprocessor."
            )
        return self._tp

    def run(self):
        r"""The entry point for starting a training process."""
        self.initialize()
        self.train()

    @abstractmethod
    def create_tp_config(self) -> Dict:
        r"""Users should overwrite this method to provide a concrete train
        preprocessor config to create an object of
        :class:`~forte.train_preprocessor.TrainPreprocessor`.
        An example config is given in the example above.

        Please refer to :meth:`default_configs` in class
        :class:`~forte.train_preprocessor.TrainPreprocessor` for detailed
        specification of each options in the config.
        """
        raise NotImplementedError

    @abstractmethod
    def create_pack_iterator(self) -> Iterator[DataPack]:
        r"""Users should overwrite this method to provide an iterator of data,
        in DataPack formats:class:`~forte.data.data_pack.DataPack`.
        This iterator will be used to produce each input data pack consumed
        for training. Typically, users can create a reader of type
        :class:`~forte.data.readers.base_reader.BaseReader`. The reader can be
        wrapped as an iterator of data pack via forte pipeline system. Please
        refer to the above example for how to create this.
        """
        raise NotImplementedError

    @abstractmethod
    def train(self):
        r"""Users should overwrite this method to provide the detail logic of
        doing the training (forward and backward processing). Users can use the
        :meth:`get_train_batch_iterator` in class
        :class:`~forte.train_preprocessor.TrainPreprocessor` to get an iterator
        of pre-processed batch of data. Please refer to that method for details.
        An example is also provided above.
        """
        raise NotImplementedError

    def save(self, *args: Any, **kwargs: Any):
        r"""Save the training states to disk for the usage of later
        predicting phase. The default training states is the request inside
        TrainPreprocessor. Please refer to :meth:`request` in class
        :class:`~forte.train_preprocessor.TrainPreprocessor` for details.
        Typically users do not need to overwrite this method as default saved
        training state is enough for predicting usage. But users can also
        overwrite this method to achieve special purpose.
        """

        # Check arg type. Default behavior only supports str as args[0] which
        # is considered as a disk file path.
        if not isinstance(args[0], str):
            raise ValueError(
                "Do not support input args: {} and kwargs: {}".format(
                    args, kwargs
                )
            )

        file_path = args[0]

        if not isinstance(self.train_preprocessor, TrainPreprocessor):
            raise ValueError(
                "Invalid TrainPreprocessor type: {}".format(
                    self.train_preprocessor
                )
            )

        request: Dict = self.train_preprocessor.request

        with open(file_path, "wb") as f:
            pickle.dump(request, f)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional, List

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.base_reader import BaseReader
from forte.evaluation.base.base_evaluator import Evaluator
from forte.pipeline import Pipeline
from forte.processors.base import BaseProcessor
from forte.trainer.base import BaseTrainer

logger = logging.getLogger(__name__)


class TrainPipeline:
    def __init__(
        self,
        train_reader: BaseReader,
        trainer: BaseTrainer,
        dev_reader: BaseReader,
        configs: Config,
        preprocessors: Optional[List[BaseProcessor]] = None,
        evaluator: Optional[Evaluator] = None,
        predictor: Optional[BaseProcessor] = None,
    ):
        self.resource = Resources()
        self.configs = configs

        train_reader.initialize(self.resource, self.configs.reader)

        if preprocessors is not None:
            for p in preprocessors:
                p.initialize(
                    resources=self.resource, configs=configs.preprocessor
                )
            self.preprocessors = preprocessors
        else:
            self.preprocessors = []

        self.train_reader = train_reader
        self.dev_reader = dev_reader
        self.trainer = trainer

        if predictor is not None:
            self.predictor = predictor

        if evaluator is not None:
            self.evaluator = evaluator
            self.evaluator.initialize(self.resource, self.configs.evaluator)

    def run(self):
        logging.info("Preparing the pipeline")
        self.prepare()
        logging.info("Initializing the trainer")

        # initialize the pipeline after prepare step, since prepare will update
        # the resources
        self.trainer.initialize(self.resource, self.configs)
        if self.predictor is not None:
            logger.info("Initializing the predictor")
            self.predictor.initialize(self.resource, self.configs)

        logging.info("The pipeline is training")
        self.train()
        self.finish()

    def prepare(self):
        prepare_pl = Pipeline()
        prepare_pl.set_reader(self.train_reader)
        for p in self.preprocessors:
            prepare_pl.add(p)
        prepare_pl.run(self.configs.config_data.train_path)

    def train(self):
        epoch = 0
        while True:
            epoch += 1
            for pack in self.train_reader.iter(
                self.configs.config_data.train_path
            ):
                for instance in pack.get_data(**self.trainer.data_request()):
                    self.trainer.consume(instance)

            self.trainer.epoch_finish_action(epoch)

            if self.trainer.validation_requested():
                dev_res = self._validate(epoch)
                self.trainer.validation_done()
                self.trainer.post_validation_action(dev_res)
            if self.trainer.stop_train():
                return

            logging.info("End of epoch %d", epoch)

    def _validate(self, epoch: int):
        validation_result = {"epoch": epoch}

        if self.predictor is not None:
            for pack in self.dev_reader.iter(self.configs.config_data.val_path):
                predicted_pack = pack.view()
                self.predictor.process(predicted_pack)
                self.evaluator.consume_next(predicted_pack, pack)
            validation_result["eval"] = self.evaluator.get_result()

        if self.evaluator is not None:
            for pack in self.dev_reader.iter(
                self.configs.config_data.test_path
            ):
                predicted_pack = pack.view()
                self.predictor.process(predicted_pack)
                self.evaluator.consume_next(predicted_pack, pack)
            validation_result["test"] = self.evaluator.get_result()

        return validation_result

    def finish(self):
        self.train_reader.finish(self.resource)
        self.dev_reader.finish(self.resource)
        self.trainer.finish(self.resource)
        self.predictor.finish(self.resource)



# ***automatically_generated***
# ***source json:forte/ontology_specs/ag_news.json***
# flake8: noqa
# mypy: ignore-errors
# pylint: skip-file
"""
Automatically generated ontology ag_news. Do not change manually.
"""

from dataclasses import dataclass
from forte.data.data_pack import DataPack
from forte.data.ontology.top import Annotation

__all__ = [
    "Description",
]


@dataclass
class Description(Annotation):

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)



"""
Script to facilitate generating ontology given a root JSON config,
and cleaning a folder out of generated ontology classes.
"""
import os
import sys
import logging
import argparse
from argparse import RawTextHelpFormatter
from forte.data.ontology.ontology_code_generator import OntologyCodeGenerator

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
log = logging.getLogger(__name__)


def normalize_path(path):
    if path is None:
        return None
    return os.path.abspath(os.path.expanduser(path))


def create(args_):
    """
    Function for the `create` mode. Generates the ontology.
    Args:
        args_: parsed args for the `create` mode
    """
    spec_path = normalize_path(args_.spec)
    dest_path = normalize_path(args_.dest_path)
    spec_paths = (
        [normalize_path(config) for config in args_.spec_paths]
        if args_.spec_paths is not None
        else None
    )
    merged_path = normalize_path(args_.merged_path)
    leient_prefix = args_.lenient_prefix

    generator = OntologyCodeGenerator(spec_paths, args_.gen_all)
    if args_.no_dry_run is None:
        log.info(
            "Ontology will be generated in a temporary directory as "
            "--no_dry_run is not specified by the user."
        )
        args_.no_dry_run = False

    if leient_prefix:
        log.info("Will not enforce prefix check.")

    is_dry_run = not args_.no_dry_run
    include_init = not args_.exclude_init
    generated_folder = generator.generate(
        spec_path,
        dest_path,
        is_dry_run,
        include_init,
        merged_path,
        leient_prefix,
    )
    log.info("Ontology generated in the directory %s.", generated_folder)


def clean(args_):
    """
    Function for the `clean` mode. Cleans the given directory of generated
    files.
    Args:
        args_: parsed args for the `clean` mode
    """
    dir_ = normalize_path(args_.dir)
    generator = OntologyCodeGenerator()
    is_empty, del_dir = generator.cleanup_generated_ontology(dir_, args_.force)
    if not is_empty:
        log.info("Directory %s not empty, cannot delete completely.", dir_)
    else:
        log.info("Directory %s deleted.", dir_)
    if not args_.force:
        log.info("Deleted files moved to %s.", del_dir)


class OntologyGenerationParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("Error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def main():
    create_example = (
        "generate_ontology create --spec "
        "ontology_specs/example/pet_shop.json"
    )
    clean_example = "generate_ontology clean --dir generated-files"

    file_description = "\n".join(
        [
            "Utility to automatically generate or create Python classes given "
            "the ontology.",
            "\n*create*: Generate ontology given a JSON specification.",
            f"Example: {create_example} --no_dry_run\n",
            "*clean*: Clean a folder of generated Python classes.",
            f"Example: {clean_example}",
        ]
    )

    parser = OntologyGenerationParser(
        description=file_description, formatter_class=RawTextHelpFormatter
    )
    subs = parser.add_subparsers(
        title="required commands", help="Select one of:"
    )

    # Parser for creating the ontology.
    create_parser = subs.add_parser("create", help="to create new ontology")
    create_parser.add_argument(
        "-i",
        "--spec",
        type=str,
        required=True,
        help="The main input JSON specification.",
    )

    create_parser.add_argument(
        "-r",
        "--no_dry_run",
        required=False,
        default=None,
        action="store_true",
        help="Generates the package tree in a temporary "
        "directory if true, ignores the argument "
        "`--dest_path`",
    )

    create_parser.add_argument(
        "-o",
        "--dest_path",
        type=str,
        required=False,
        default=os.getcwd(),
        help="Destination directory provided by the user"
        ". Only used when --no_dry_run is "
        "specified. The"
        " default directory is the current working "
        "directory.",
    )

    create_parser.add_argument(
        "-s",
        "--spec_paths",
        type=str,
        nargs="*",
        required=False,
        default=None,
        help="Paths in which the root and imported "
        "spec files are to be searched.",
    )

    create_parser.add_argument(
        "-m",
        "--merged_path",
        type=str,
        required=False,
        default=None,
        help="The destination file path for the merged" "file path.",
    )

    create_parser.add_argument(
        "-e",
        "--exclude_init",
        required=False,
        default=None,
        action="store_true",
        help="Excludes generation of `__init__.py` files"
        " in the already existing directories, if"
        "`__init__.py` not already present.",
    )

    create_parser.add_argument(
        "-a",
        "--gen_all",
        required=False,
        default=False,
        action="store_true",
        help="If True, will generate all the ontology,"
        "including the existing ones shipped with "
        "Forte.",
    )

    create_parser.add_argument(
        "-l",
        "--lenient_prefix",
        required=False,
        default=False,
        action="store_true",
        help="If True, will not enforce prefix check.",
    )

    create_parser.add_argument(
        "-n",
        "--namespace_depth",
        type=int,
        required=False,
        default=0,
        help="set an integer argument namespace_depth"
        "to allow customized number of levels of namespace packaging."
        "The generation of __init__.py for all the"
        "directory levels above namespace_depth will be disabled."
        "Default value is set to 0 to disable namespace packaging"
        "When namespace_depth<=0, namespace packaging will be disabled"
        "and __init__.py will be included in all directory levels",
    )

    create_parser.set_defaults(func=create)

    # Parsing for cleaning.
    clean_parser = subs.add_parser("clean", help="to clean up ontologies")

    clean_parser.add_argument(
        "-d",
        "--dir",
        type=str,
        required=True,
        help="Generated files to be cleaned from the " "directory path.",
    )

    clean_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="If true, skips the interactive deleting of "
        "folders. Use with caution.",
    )

    clean_parser.set_defaults(func=clean)

    options = parser.parse_args()
    options_func = getattr(options, "func", None)
    if options_func is not None:
        options.func(options)
    else:
        sys.stderr.write("Error: %s\n" % "wrong usage of the script.")
        OntologyGenerationParser().print_help()
        sys.exit(2)


if __name__ == "__main__":
    main()



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Create a few common functions that are frequently used to interact with
built-in ontology data types.
"""
from typing import Optional

from forte.data.data_pack import DataPack
from ft.onto.base_ontology import Utterance

__all__ = [
    "create_utterance",
    "get_last_utterance",
]


def get_last_utterance(
    input_pack: DataPack, target_speaker: str
) -> Optional[Utterance]:
    """
    Get the last utterance from a particular speaker. An utterance is an entry
    of type :class:`~ft.onto.base_ontology.Utterance`

    Args:
        input_pack: The data pack to find utterances.
        target_speaker: The name of the target speaker.

    Returns:
        The last Utterance from the speaker if found, None otherwise.
    """
    utterance: Optional[Utterance] = None
    u: Utterance
    for u in input_pack.get(Utterance):
        if u.speaker == target_speaker:
            utterance = u

    return utterance


def create_utterance(input_pack: DataPack, text: str, speaker: str):
    """
    Create an utterance in the datapack. This is composed of two steps:

        1. Append the utterance text to the data pack.
        2. Create :class:`~ft.onto.base_ontology.Utterance` entry on the text.
        3. Set the speaker of the utterance to the provided `speaker`.

    Args:
        input_pack: The data pack to add utterance into.
        text: The text of the utterance.
        speaker: The speaker name to be associated with the utterance.

    """
    input_pack.set_text(input_pack.text + "\n" + text)

    u = Utterance(
        input_pack, len(input_pack.text) - len(text), len(input_pack.text)
    )
    u.speaker = speaker



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Subword Tokenizer"""

__all__ = [
    "SubwordTokenizer",
]

from typing import List, Tuple, Iterator, Dict, Set


from forte.common import Resources
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.processors.base import PackProcessor
from forte.utils.utils import DiffAligner
from forte.utils import create_import_error_msg
from ft.onto.base_ontology import Subword


try:
    from texar.torch.data.tokenizers.bert_tokenizer import BERTTokenizer
except ImportError as err1:
    raise ImportError(
        create_import_error_msg("texar-pytorch", "nlp", "NLP support")
    ) from err1

# This should probably be named as `BertTokenizer`.
class SubwordTokenizer(PackProcessor):
    """
    Subword Tokenizer using pretrained Bert model.
    """

    def __init__(self):
        super().__init__()

        self.tokenizer: BERTTokenizer = None
        self.aligner: DiffAligner = None
        self.__do_lower_case = True

    # pylint: disable=attribute-defined-outside-init,unused-argument
    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        if not self.configs.tokenizer_configs.pretrained_model_name:
            raise ValueError("Please specify a pretrained bert model")
        self.tokenizer = BERTTokenizer(
            cache_dir=None,
            hparams=self.configs.tokenizer_configs,
        )
        self.aligner = DiffAligner()
        self.__do_lower_case = self.configs.tokenizer_configs.do_lower_case

    def _process(self, input_pack: DataPack):
        assert self.tokenizer is not None
        assert self.aligner is not None

        if self.configs.token_source is not None:
            # Use provided token source.
            token: Annotation
            for token in input_pack.get(self.configs.token_source):
                assert isinstance(token, Annotation)
                self.__add_subwords(
                    input_pack,
                    token.text  # type: ignore
                    if self.__do_lower_case
                    else token.text,  # type: ignore
                    token.begin,  # type: ignore
                )
        elif self.configs.segment_unit is not None:
            # If token source not provide, try to use provided segments.
            segment: Annotation
            for segment in input_pack.get(self.configs.segment_unit):
                self._segment(
                    input_pack, segment.text, segment.begin  # type: ignore
                )
        else:
            # Use the whole data pack, maybe less efficient in some cases.
            self._segment(input_pack, input_pack.text, 0)

    def _word_tokenization(
        self, text: str
    ) -> Iterator[Tuple[str, Tuple[int, int]]]:
        """
        This function should tokenize the text and return the tokenization
        results in the form of a word and the span of each word. A span is the
        begin and end of this word, indexed from 0, and end = begin + length
        of the word.

        By default, this calls the Texar's BasicTokenizer and then align the
        result back. You can implement this function if you prefer a
        different tokenizer.

        Args:
            text: Input text to be tokenized.

        Returns:
            A iterator of tokenization result in the form of triplets of
            (word, (begin, end)).
        """
        basic_tokens: List[str] = self.tokenizer.basic_tokenizer.tokenize(
            text, never_split=self.tokenizer.all_special_tokens
        )
        token_spans = self.aligner.align_with_segments(text, basic_tokens)

        for t, span in zip(basic_tokens, token_spans):
            if span is not None:
                yield t, span

    def _segment(self, pack: DataPack, text: str, segment_offset: int):
        if self.tokenizer.do_basic_tokenize:
            for token, (token_begin, _) in self._word_tokenization(text):
                assert token is not None
                self.__add_subwords(pack, text, token_begin + segment_offset)
        else:
            self.__add_subwords(pack, text, segment_offset)

    def __add_subwords(self, pack: DataPack, text: str, text_offset: int):
        if self.__do_lower_case:
            lower_text = text.lower()
            # See this https://bugs.python.org/issue17252 to understand why this
            # is checked here. tl;dr sometimes lower casing special unicode
            # string will result in a change of length due to unicode NFD.
            if len(lower_text) == len(text):
                text = text.lower()

        for (
            subword,
            begin,
            end,
        ) in self.tokenizer.wordpiece_tokenizer.tokenize_with_span(text):
            subword_token = Subword(
                pack, begin + text_offset, end + text_offset
            )
            if subword == self.tokenizer.wordpiece_tokenizer.unk_token:
                subword_token.is_unk = True
            subword_token.is_first_segment = not subword.startswith("##")
            # pylint: disable=protected-access
            subword_token.vocab_id = self.tokenizer._map_token_to_id(subword)

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of current processor
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the data pack storing type records needed
                in for consistency checking.

        Returns:
            None
        """
        record_meta["ft.onto.base_ontology.Subword"] = {
            "is_unk",
            "is_first_segment",
            "vocab_id",
        }

    def expected_types_and_attributes(self) -> Dict[str, Set[str]]:
        r"""Method to add expected type for current processor input which
        would be checked before running the processor if
        the pipeline is initialized with
        `enforce_consistency=True` or
        :meth:`~forte.pipeline.Pipeline.enforce_consistency` was enabled for
        the pipeline.
        """
        expected_types: Dict[str, Set[str]] = {}
        if self.configs.token_source is not None:
            expected_types[self.configs.token_source] = set()
        elif self.configs.segment_unit is not None:
            expected_types[self.configs.segment_unit] = set()
        return expected_types

    @classmethod
    def default_configs(cls):
        """Returns the configuration with default values.

        Here:

        - `tokenizer_configs` contains all default hyper-parameters in
          :class:`~texar.torch.data.tokenizer.bert_tokenizer.BERTTokenizer`,
          this processor will pass on all the configurations to the
          tokenizer to create the tokenizer instance.

        - `segment_unit` contains an Annotation entry type used to split the
          text into smaller units. For example, setting this to
          `ft.onto.base_ontology.Sentence` will make this tokenizer do
          tokenization on a sentence base, which could be more efficient
          when the alignment is used.

        - `token_source` contains entry name of where the tokens come from.
          For example, setting this to `ft.onto.base_ontology.Token` will
          make this tokenizer split the sub-word based on this token. The
          default value will use `ft.onto.base_ontology.Token`. If this
          value is set to None, then it will use `word_tokenization`
          function of this class to do tokenization.

        Note that if `segment_unit` or `token_source` is provided, the
        :meth:`~forte.processors.base.base_processor.BaseProcessor.check_record`
        will check if certain types are written before this processor.

        Returns: Default configuration value for the tokenizer.

        """
        return {
            "tokenizer_configs": BERTTokenizer.default_hparams(),
            "segment_unit": None,
            "token_source": "ft.onto.base_ontology.Token",
            "@no_typecheck": "token_source",
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the BaseTrainer, which is a core pipeline component. To
implement new trainer in Forte, one need to inherit this trainer.
"""

from abc import abstractmethod
from typing import Dict, Iterator

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.pipeline_component import PipelineComponent


class BaseTrainer(PipelineComponent):
    def __init__(self):  # pylint: disable=unused-argument
        """Create a machine learning trainer. This abstract class defines the
        skeleton of a trainer, which correspond to events and actions during a
        training phase. New trainers need to inherit and implement the
        functions.
        """
        super().__init__()
        self._stop_train = False
        self._validation_requested = False

    @abstractmethod
    def initialize(self, resources: Resources, configs: Config):
        """
        The training pipeline will run this initialization method during
        the initialization phase and send resources in as parameters.

        Args:
            resources: The resources shared in the pipeline
            configs: configuration object for this trainer.

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def data_request(self) -> Dict:
        """
        Return a request to a :class:`DataPack <forte.data.data_pack.DataPack>`.
        The trainer will use the data returned from the request for training.

        See :func:`get_data <forte.data.data_pack.DataPack.get_data>` for the
        request details.

        Returns:

        """
        pass

    @abstractmethod
    def consume(self, instance: Dict):
        """
        Consumes the next data instance. The trainer will decide the appropriate
        actions for the instance (e.g. build a batch, or train immediately).
        The instance is constructed by querying :func:`data_request`.

        Args:
            instance: The training data instance.

        Returns:

        """
        # consume the instance
        raise NotImplementedError

    @abstractmethod
    def post_validation_action(self, dev_res):
        """
        This method will be called when validation is done. Typical usage
        includes saving the development results.

        Args:
            dev_res: The evaluation results for development data.

        Returns:

        """
        pass

    @abstractmethod
    def get_loss(self, instances: Iterator[Dict]):
        """
        Compute loss based on the instances.

        Args:
            instances: An iterator containing the data instances.

        Returns:

        """
        raise NotImplementedError

    def update_resource(self):
        r"""Update the resource after every epoch which can be consumed by the
        predictor
        """
        pass

    def pack_finish_action(self, pack_count: int):
        """
        This function will be called by the pipeline when one pack is
        finished.

        Args:
            pack_count:

        Returns:

        """
        pass

    def epoch_finish_action(self, epoch_num: int):
        """
        This function will be called by the pipeline when one epoch is
        finished. For example, the trainer can call :meth:`request_stop_train`
        when the number of epoch reaches a predefined value.

        Args:
            epoch_num: The current number of epoch.

        Returns:

        """
        pass

    def request_eval(self):
        """
        The trainer should call this method to inform the pipeline to
        conduct evaluation.

        Returns:

        """
        self._validation_requested = True

    def request_stop_train(self):
        """
        The trainer should call this method to inform the pipeline to stop
        training.
        Returns:

        """
        self._stop_train = True

    def validation_done(self):
        """
        Used only by the pipeline to close the validation request.

        Returns:

        """
        self._validation_requested = False

    def validation_requested(self) -> bool:
        """
        Used only by the pipeline to check whether the trainer has made
        the validation request.

        Returns: True if the validation request is submitted and not completed.
        """
        return self._validation_requested

    def stop_train(self) -> bool:
        """
        Used only by the pipeline to check if the trainer decided to stop
        training.

        Returns: True if the trainer decided to stop.
        """
        return self._stop_train



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The processors that process data in batch.
"""
from abc import abstractmethod, ABC
from typing import List, Dict, Optional, Any

from forte.common import Resources, ProcessorConfigError
from forte.common.configuration import Config
from forte.data import slice_batch, BaseExtractor
from forte.data.base_pack import PackType
from forte.data.batchers import (
    ProcessingBatcher,
    FixedSizeRequestDataPackBatcher,
    FixedSizeDataPackBatcherWithExtractor,
)
from forte.data.converter import Converter
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology.top import Annotation
from forte.processors.base.base_processor import BaseProcessor
from forte.utils import extractor_utils

__all__ = [
    "BaseBatchProcessor",
    "PackingBatchProcessor",
    "Predictor",
    "MultiPackBatchProcessor",
    "RequestPackingProcessor",
    "FixedSizeBatchProcessor",
]

from forte.utils.extractor_utils import (
    parse_feature_extractors,
)


class BaseBatchProcessor(BaseProcessor[PackType], ABC):
    r"""The base class of processors that process data in batch. This processor
    enables easy data batching via analyze the context and data objects. The
    context defines the scope of analysis of a particular task.

    For example, in dependency parsing, the context is normally a sentence,
    in entity coreference, the context is normally a document. The processor
    will create data batches relative to the context.

    Key fields in this processor:

        - batcher: The processing batcher used for this processor. The batcher
          will also keep track of the relation between the pack and the batch
          data.

        - use_coverage_index: If true, the index will be built based on the
          requests.
    """

    def __init__(self):
        super().__init__()
        self._batcher: Optional[ProcessingBatcher] = None
        self.use_coverage_index = False

    def initialize(self, resources: Resources, configs: Optional[Config]):
        super().initialize(resources, configs)

        assert configs is not None
        try:
            self.batcher.initialize(configs.batcher)
        except AttributeError as e:
            raise ProcessorConfigError(
                "Error in handling batcher config, please check the "
                "config of the batcher to see they are correct."
            ) from e

    @property
    def batcher(self) -> ProcessingBatcher:
        if self._batcher is None:
            self._batcher = self.define_batcher()
        return self._batcher

    def _process(self, input_pack: PackType):
        r"""In batch processors, all data are processed in batches. This
        function is already implemented to convert data in to batches. Users do
        not need to implement this function, but should instead implement
        ``_process_batch``.

        but should instead implement ``predict``, which computes results from
        batches, and ``pack_all``, which convert the batch results back to
        datapacks.

        Args:
            input_pack: The next input pack to be fed in.
        """
        if self.use_coverage_index:
            self._prepare_coverage_index(input_pack)

        for packs, instances, batch in self.batcher.get_batch(input_pack):
            self._process_batch(packs, instances, batch)

    def flush(self):
        for packs, instances, batch in self.batcher.flush():
            self._process_batch(packs, instances, batch)

    @abstractmethod
    def _process_batch(
        self,
        packs: List[PackType],
        contexts: List[Optional[Annotation]],
        batched_data: Dict,
    ):
        """
        Users can implement this function to process the extracted batch
        data. This is suitable to be implemented if one do not need to add
        information back to the data pack. Otherwise, It is advised to implement
        :class:`~forte/processors/base.batch_processor.PackingBatchProcessor`
        instead of this one since PackingBatchProcessor provides helper
        functions to re-align the output with the data pack.

        Args:
            packs: The list of data packs corresponding to the batch.
            contexts: The list of context corresponding to the batch. It
                could contain `None` values, which means the the whole data
                pack will be used as contexts.
            batched_data: The data batch to be process.

        Returns:

        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """Defines the default configs for batching processor."""
        return {}

    def _prepare_coverage_index(self, input_pack: PackType):
        """
        Build the coverage index for ``input_pack``. After building, querying
          data in this pack will become more efficient.

        The index will be built based on the `input_info` field.

        Args:
            input_pack: The pack to be built.

        Returns:

        """
        pass

    @classmethod
    @abstractmethod
    def define_batcher(cls) -> ProcessingBatcher:
        r"""Define a specific batcher for this processor.
        Single pack
        :class:`~forte.processors.base.batch_processor.BaseBatchProcessor`
        initialize the batcher to be a
        :class:`~forte.data.batchers.ProcessingBatcher`.
        And
        :class:`~forte.processors.base.batch_processor.MultiPackBatchProcessor`
        initialize the batcher to be a
        :class:`~forte.processors.base.batch_processor.MultiPackBatchProcessor`
        .
        """
        raise NotImplementedError


class PackingBatchProcessor(BaseBatchProcessor[PackType], ABC):
    """
    This class extends the BaseBatchProcessor class and provide additional
    utilities to align and pack the extracted results back to the data pack.

    To implement this processor, one need to implement:
    1. The `predict` function that make predictions for each input data batch.
    2. The `pack` function that add the prediction value back to the data pack.

    Users that implement the processor only have to concern about a single
    batch, the alignment between the data batch and the data pack will be
    maintained by the system.
    """

    def _process_batch(
        self,
        packs: List[PackType],
        contexts: List[Optional[Annotation]],
        batched_data: Dict,
    ):
        """
        Implement the function into a prediction and pack step, users don't
        need to implement this function and should implement `predict` and
        `pack` instead.

        Args:
            packs (List[PackType]):  List of data packs each batch comes from.
            batched_data (Dict): The batched data.
            contexts (Optional[List[Annotation]]): List of the data contexts
                where the data comes from.

        Returns:

        """
        pred = self.predict(batched_data)
        self.pack_all(packs, contexts, pred)

    def predict(self, data_batch: Dict) -> Dict[str, List[Any]]:
        r"""The function that task processors should implement. Make
        predictions for the input ``data_batch``.

        Args:
              data_batch (dict): A batch of instances in our ``dict`` format.

        Returns:
              The prediction results in dictionary form.
        """
        raise NotImplementedError

    def pack(
        self,
        pack: PackType,
        predict_results: Dict[str, Any],
        context: Optional[Annotation] = None,
    ):
        r"""The function that task processors should implement. It is the
        custom function on how to add the predicted output back to the data
        pack.

        Args:
            pack: The pack to add entries or fields to.
            predict_results: The prediction results returned by
                :meth:`~forte.processors.base.batch_processor
                .BaseBatchProcessor.predict`.
                This processor will add these results to the provided `pack`
                as entry and attributes.
            context: The context entry that the
                prediction is performed, and the pack operation should
                be performed related to this range annotation. If None,
                then we consider the whole data pack is used as the context.
        """
        raise NotImplementedError

    def pack_all(
        self,
        packs: List[PackType],
        contexts: List[Optional[Annotation]],
        output_dict: Dict[str, List[Any]],
    ):
        r"""
        Pack the prediction results contained in the `output_dict` back to the
        corresponding packs.

        Args:
            packs: The list of data packs corresponding to the output batches.
            contexts: The list of contexts corresponding to the output batches.
            output_dict: Stores the output in a specific format. The keys
                are string names that specify data. The value is a list of
                data in the shape of (batch_size, Any). There might be
                additional structures inside `Any` as specific
                implementation choices.
        """
        # Group the same pack and context into the same segments.

        # The list of (pack, context) tuple.
        pack_context_pool = []
        # Store the segments of the pack context, the len of this list should
        # be the number of segments, and the value indicates the length of
        # the segment. These will be used to slice the data batch.
        segment_lengths = []

        # Note that this will work if the elements in `contexts` are all
        # None, which means they are all the same, and will have the correct
        # behavior.
        prev_pack_context = None
        for pack_context_i in zip(packs, contexts):
            if pack_context_i != prev_pack_context:
                segment_lengths.append(1)
                prev_pack_context = pack_context_i
                pack_context_pool.append(pack_context_i)
            else:
                segment_lengths[-1] += 1

        start = 0

        for i, (pack_i, context) in enumerate(pack_context_pool):
            # The slice should correspond to the portion of the data batch
            # that should be assigned to these pack and context.
            output_dict_i = slice_batch(output_dict, start, segment_lengths[i])
            self.pack(pack_i, output_dict_i, context)
            start += segment_lengths[i]
            pack_i.add_all_remaining_entries()


class FixedSizeBatchProcessor(PackingBatchProcessor[DataPack], ABC):
    """
    A processor that implements the packing batch processor, using a fixed
    size batcher :class:`~forte.data.batchers.FixedSizeDataPackBatcher`
    """

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """Defines the default configs for batching processor."""
        return {
            "batcher": {
                "batch_size": 4,
            },
        }


class RequestPackingProcessor(PackingBatchProcessor[DataPack], ABC):
    """
    A processor that implements the packing batch processor, using a
    variation of the fixed size batcher
    :class:`~forte.data.batchers.FixedSizeRequestDataPackBatcher`,
    which will use `DataPack.get_data` function with the`context_type`
    and `requests` parameters.
    """

    @classmethod
    def define_batcher(cls) -> ProcessingBatcher:
        return FixedSizeRequestDataPackBatcher()

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """Defines the default configs for batching processor."""
        return {
            "batcher": {"batch_size": 4, "context_type": None, "requests": {}},
        }


class Predictor(PackingBatchProcessor[PackType]):
    r"""
    `Predictor` is a special type of batch processor that uses
    :class:`~forte.data.BaseExtractor` to collect features from data packs, and
    also uses Extractors to write the prediction back.

    `Predictor` implements the `PackingBatchProcessor` class, and implements
    the `predict` and `pack` function using the extractors.
    """

    def __init__(self):
        super().__init__()
        self.model = None
        self.do_eval = False
        self._request: Dict = {}
        self._request_ready: bool = False

        self._batcher: Optional[FixedSizeDataPackBatcherWithExtractor] = None

    def add_extractor(
        self,
        name: str,
        extractor: BaseExtractor,
        is_input: bool,
        converter: Optional[Converter] = None,
    ):
        """
        Extractors can be added to the preprocessor directly via this
        method.

        Args:
            name: The name/identifier of this extractor, the name should be
              different between different extractors.
            extractor: The extractor instance to be added.
            is_input: Whether this extractor will be used as input or output.
            converter:  The converter instance to be applied after running
              the extractor.

        Returns:
            None
        """
        extractor_utils.add_extractor(
            self._request, name, extractor, is_input, converter
        )

    def set_feature_requests(self, request: Dict):
        self._request = request
        self._request_ready = True

    def deactivate_request(self):
        self._request_ready = False

    @classmethod
    def define_batcher(cls) -> ProcessingBatcher:
        return FixedSizeDataPackBatcherWithExtractor()

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {
            "feature_scheme": None,
            "context_type": None,
            "batcher": cls.define_batcher().default_configs(),
            "do_eval": False,
        }

    def initialize(self, resources: Resources, configs: Config):
        # Populate the _request. The self._request_ready help avoid parsing
        # the feature scheme multiple times during `initialize`.
        if not self._request_ready:
            for key, value in configs.items():
                if key == "feature_scheme":
                    self._request["schemes"] = parse_feature_extractors(
                        configs.feature_scheme
                    )
                else:
                    self._request[key] = value
            self._request_ready = True

        batcher_config = configs.batcher
        # Assign context type from here to make sure batcher is using the
        # same context type as predictor.
        batcher_context = configs["batcher"].get("context_type", None)
        if (
            batcher_context is None
            or batcher_context == self._request["context_type"]
        ):
            batcher_config.context_type = self._request["context_type"]
        else:
            raise ProcessorConfigError(
                "The 'context_type' configuration value should be the same "
                "for the processor and the batcher, now for the processor the "
                f"value is {self._request['context_type']} and for the "
                f"batcher the value is {batcher_context}. It is also fine if "
                f"this value for batch config is left empty."
            )
        self.do_eval = configs.do_eval

        # This needs to be called later since batcher config needs to be loaded.
        super().initialize(resources, configs)
        for tag, scheme in self._request["schemes"].items():
            # Add input feature to the batcher.
            if scheme["type"] == extractor_utils.DATA_INPUT:
                self.batcher.add_feature_scheme(tag, scheme)  # type: ignore

    def load(self, model):
        self.model = model

    def pack(
        self,
        pack: PackType,
        predict_results: Dict,
        context: Optional[Annotation] = None,
    ):
        for tag, batched_predictions in predict_results.items():
            # preds contains batched results.
            if self.do_eval:
                self.__extractor(tag).pre_evaluation_action(pack, context)
            for prediction in batched_predictions:
                self.__extractor(tag).add_to_pack(pack, prediction, context)
        pack.add_all_remaining_entries()

    def __extractor(self, tag_name: str):
        return self._request["schemes"][tag_name]["extractor"]

    def predict(self, data_batch: Dict) -> Dict:
        r"""The function that task processors should implement. Make
        predictions for the input ``data_batch``.

        Args:
              data_batch: A batch of instances in our ``dict`` format.

        Returns:
              The prediction results in dict datasets.
        """
        raise NotImplementedError()


class MultiPackBatchProcessor(BaseBatchProcessor[MultiPack], ABC):
    r"""This class defines the base batch processor for `MultiPack`."""

    def __init__(self):
        super().__init__()
        self.input_pack_name = None
        # TODO multi pack batcher need to be further implemented.



import json
from typing import List
from forte.utils import create_import_error_msg
from forte.models.srl.data import RawExample, Example, SRLSpan

try:
    import torch
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "models")
    ) from e


try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "models", "Texar model support"
        )
    ) from e


class SRLSpanData(tx.data.DatasetBase[str, Example]):
    def __init__(self, path: str, vocab: tx.data.Vocab, hparams):
        source = tx.data.TextLineDataSource(path)
        self._vocab = vocab
        super().__init__(source, hparams)

    def process(self, raw_example: str) -> Example:
        raw: RawExample = json.loads(raw_example)
        assert len(raw["sentences"]) == 1
        sentence = raw["sentences"][0]
        example: Example = {
            "text": sentence,
            "text_ids": self._vocab.map_tokens_to_ids_py(sentence),
            "srl": [SRLSpan(*items) for items in raw["srl"][0]],
        }
        return example

    def collate(self, examples: List[Example]) -> tx.data.Batch:
        sentences = [ex["text"] for ex in examples]
        tokens, length = tx.data.padded_batch(
            [ex["text_ids"] for ex in examples]
        )
        srl = [ex["srl"] for ex in examples]
        return tx.data.Batch(
            len(examples),
            srl=srl,
            text=sentences,
            text_ids=torch.from_numpy(tokens).to(self.device),
            length=torch.tensor(length).to(self.device),
        )



# ***automatically_generated***
# ***source json:../testing/metric.json***
# flake8: noqa
# mypy: ignore-errors
# pylint: skip-file
"""
Automatically generated ontology metric. Do not change manually.
"""

from dataclasses import dataclass
from forte.data.data_pack import DataPack
from forte.data.ontology.core import FNdArray
from forte.data.ontology.top import Generics
from typing import Optional

__all__ = [
    "Metric",
    "SingleMetric",
    "NdMetric",
]


@dataclass
class Metric(Generics):
    """
    A base metric entity, all metric entities should inherit from it.
    Attributes:
        metric_name (Optional[str]):
    """

    metric_name: Optional[str]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.metric_name: Optional[str] = None


@dataclass
class SingleMetric(Metric):
    """
    A single metric entity, used to present a metric of one float (e.g. accuracy).
    Attributes:
        value (Optional[float]):
    """

    value: Optional[float]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.value: Optional[float] = None


@dataclass
class NdMetric(Metric):
    """
    A N-dimensional metric entity, used to present a metric of N-d array (e.g. confusion matrix).
    Attributes:
        value (FNdArray):
    """

    value: FNdArray

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.value: FNdArray = FNdArray(shape=None, dtype='float')



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Evaluator for Conll03 NER tag."""
import os
from typing import Dict
from pathlib import Path

from ft.onto.base_ontology import Sentence
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.evaluation.base import Evaluator
from forte.utils import get_class
from forte.datasets.conll.conll_utils import write_tokens_to_file


class CoNLLNEREvaluator(Evaluator):
    def __init__(self):
        super().__init__()
        self.output_file = "tmp_eval.txt"
        self.score_file = "tmp_eval.score"
        self.scores: Dict[str, float] = {}
        if os.path.isfile(self.output_file):
            os.remove(self.output_file)

    def initialize(self, resources: Resources, configs: Config):
        # pylint: disable=attribute-defined-outside-init,unused-argument
        r"""Initialize the evaluator with `resources` and `configs`.
        This method is called by the pipeline during the initialization.

        Args:
            resources: An object of class
                :class:`~forte.common.Resources` that holds references to
                objects that can be shared throughout the pipeline.
            configs: A configuration to initialize the
                evaluator. This evaluator is expected to hold the
                following (key, value) pairs
                - `"entry_type"` (str): The entry to be evaluated.
                - `"tagging_unit"` (str): The tagging unit that the evaluation
                is performed on. e.g. `"ft.onto.base_ontology.Sentence"`
                - `"attribute"` (str): The attribute of the entry to be
                evaluated.

        """
        super().initialize(resources, configs)
        self.entry_type = get_class(configs.entry_type)
        self.tagging_unit = get_class(configs.tagging_unit)
        self.attribute = configs.attribute
        self.__eval_script = configs.eval_script

    @classmethod
    def default_configs(cls):
        config = super().default_configs()
        config.update(
            {
                "entry_type": None,
                "tagging_unit": None,
                "attribute": "",
                "eval_script": str(
                    Path(os.path.abspath(__file__)).parents[1]
                    / "utils/eval_scripts/conll03eval.v2"
                ),
            }
        )
        return config

    def consume_next(self, pred_pack: DataPack, ref_pack: DataPack):
        refer_getdata_args = {
            "context_type": Sentence,
            "request": {
                self.tagging_unit: {"fields": ["ner"]},
                self.entry_type: {
                    "fields": [self.attribute],
                },
            },
        }
        write_tokens_to_file(
            pred_pack,
            ref_pack,
            refer_getdata_args,
            self.tagging_unit,
            self.entry_type,
            self.attribute,
            self.output_file,
        )

    def get_result(self) -> Dict:
        eval_call = (
            f"perl {self.__eval_script} < {self.output_file} > "
            f"{self.score_file}"
        )

        call_return = os.system(eval_call)
        if call_return == 0:
            with open(self.score_file, "r", encoding="utf-8") as fin:
                fin.readline()
                line = fin.readline()
                fields = line.split(";")
                acc = float(fields[0].split(":")[1].strip()[:-1])
                precision = float(fields[1].split(":")[1].strip()[:-1])
                recall = float(fields[2].split(":")[1].strip()[:-1])
                f_1 = float(fields[3].split(":")[1].strip())

            self.scores = {
                "accuracy": acc,
                "precision": precision,
                "recall": recall,
                "f1": f_1,
            }
            return self.scores
        else:
            raise RuntimeError(
                f"Error running eval script, return code is {call_return} "
                f"when running the command {eval_call}"
            )



# Copyright 2022 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SingleAnnotationAugmentOp is the extension of the BaseDataAugmentationOp.
The aim of this Op is to increase ease of use and reduce freedom of
operations that can be done on a data pack. This Op only allows the
implementation of those augmentation algorithms that process one
annotation at a time. This annotation can be one Token or even one
Sentence. But once declares, the augmentation will happen only
on the specified annotations.
"""
from typing import (
    Tuple,
    Dict,
    Any,
)
from abc import abstractmethod
from forte.data.data_pack import DataPack
from forte.data.ontology.top import Annotation
from forte.utils.utils import get_class
from forte.processors.data_augment.algorithms.base_data_augmentation_op import (
    BaseDataAugmentationOp,
)


__all__ = ["SingleAnnotationAugmentOp"]


class SingleAnnotationAugmentOp(BaseDataAugmentationOp):
    r"""
    This class extends the `BaseDataAugmentationOp` to
    only allow augmentation of one annotation at a time.
    This operation should be used when we only want to
    augment one type of annotation in the whole data pack.
    Thus, to use this operation, the developer only needs
    to specify how a single annotation will be processed
    as a part of their augmentation method.
    We leave the :func:`single_annotation_augment` method to
    be implemented by the subclass. This function will specify
    what type of augmentation will a given annotation (of a
    predefined type) undergo.
    """

    def augment(self, data_pack: DataPack) -> bool:
        r"""
        This method is not to be modified when using
        the `SingleAnnotationAugmentOp`. This function takes
        in the augmentation logic specified by :func:`single_annotation_augment`
        method to apply it to each annotation of the specified type individually.

        Args:
            input_anno: the input annotation to be replaced.

        Returns:
            A boolean value indicating if the augmentation
            was successful (True) or unsuccessful (False).
        """
        augment_entry = get_class(self.configs["augment_entry"])
        anno: Annotation
        replaced_text: str
        is_replace: bool
        for anno in data_pack.get(augment_entry):
            is_replace, replaced_text = self.single_annotation_augment(anno)
            if is_replace:
                try:
                    _ = self.replace_annotations(anno, replaced_text)
                except ValueError:
                    return False
        return True

    @abstractmethod
    def single_annotation_augment(
        self, input_anno: Annotation
    ) -> Tuple[bool, str]:
        r"""
        This function takes in one annotation at a time and performs
        the desired augmentation on it. Through this function, one
        annotation is processed at a time. The developer needs to specify the
        logic that will be adopted to process one annotation of a given type.
        This method cannot suggest an augmentation logic which take in multiple
        annotations of the same type.

        Args:
            input_anno: The annotation that needs to be augmented.

        Returns:
            A tuple, where the first element is a boolean value indicating
            whether the augmentation happens, and the second element is the
            replaced string.
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        r"""
        Returns:
            A dictionary with the default config for this processor.
        Following are the keys for this dictionary:
            - augment_entry:
                Defines the entry the processor will augment.
                It should be a full qualified name of the entry class.
                Default value is "ft.onto.base_ontology.Token".
        """
        return {"augment_entry": "ft.onto.base_ontology.Token"}



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=attribute-defined-outside-init
import pickle
from typing import Any, Dict, Tuple
import numpy as np
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.processors.base import QueryProcessor


try:
    import torch
except ImportError as e1:
    raise ImportError(
        create_import_error_msg("torch", "ir", "Information Retrieval supports")
    ) from e1


__all__ = ["BertBasedQueryCreator"]


class BertBasedQueryCreator(QueryProcessor):
    r"""This processor searches relevant documents for a query"""

    # pylint: disable=useless-super-delegation
    def __init__(self) -> None:
        super().__init__()

    def initialize(self, resources: Resources, configs: Config):
        self.resource = resources
        self.config = configs

        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )

        try:
            from texar.torch.data import (  # pylint: disable=import-outside-toplevel
                BERTTokenizer,
            )
            from texar.torch.modules import (  # pylint: disable=import-outside-toplevel
                BERTEncoder,
            )
        except ImportError as e:
            raise ImportError(
                create_import_error_msg("texar-pytorch", "ir", "IR support")
            ) from e

        if "name" in self.config.tokenizer:
            self.tokenizer = BERTTokenizer(
                pretrained_model_name=self.config.tokenizer.name
            )

        if "name" in self.config.model:
            self.encoder = BERTEncoder(
                pretrained_model_name=self.config.model.name
            )

        else:
            self.encoder = BERTEncoder(
                pretrained_model_name=None,
                hparams={"pretrained_model_name": None},
            )
            with open(self.config.model.path, "rb") as f:
                state_dict = pickle.load(f)
            self.encoder.load_state_dict(state_dict["bert"])

        self.encoder.to(self.device)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {
            "model": {
                "path": None,
                "name": "bert-base-uncased",
            },
            "tokenizer": {"name": "bert-base-uncased"},
            "max_seq_length": 128,
            "query_pack_name": "query",
        }

    @torch.no_grad()
    def get_embeddings(self, inputs, sequence_length, segment_ids):
        output, _ = self.encoder(
            inputs=inputs,
            sequence_length=sequence_length,
            segment_ids=segment_ids,
        )
        cls_token = output[:, 0, :]

        return cls_token

    def _build_query(self, text: str) -> np.ndarray:
        (input_ids, segment_ids, input_mask,) = self.tokenizer.encode_text(
            text_a=text, max_seq_length=self.config.max_seq_length
        )
        input_ids = torch.LongTensor(input_ids).unsqueeze(0).to(self.device)
        segment_ids = torch.LongTensor(segment_ids).unsqueeze(0).to(self.device)
        input_mask = torch.LongTensor(input_mask).unsqueeze(0).to(self.device)
        sequence_length = (~(input_mask == 0)).sum(dim=1)
        query_vector = self.get_embeddings(
            inputs=input_ids,
            sequence_length=sequence_length,
            segment_ids=segment_ids,
        )
        query_vector = torch.mean(query_vector, dim=0, keepdim=True)
        query_vector = query_vector.cpu().numpy()
        return query_vector

    def _process_query(
        self, input_pack: MultiPack
    ) -> Tuple[DataPack, np.ndarray]:
        query_pack: DataPack = input_pack.get_pack(self.config.query_pack_name)
        context = [query_pack.text]

        # use context to build the query
        if "user_utterance" in input_pack.pack_names:
            user_pack = input_pack.get_pack("user_utterance")
            context.append(user_pack.text)

        if "bot_utterance" in input_pack.pack_names:
            bot_pack = input_pack.get_pack("bot_utterance")
            context.append(bot_pack.text)

        text = " ".join(context)

        query_vector = self._build_query(text=text)

        return query_pack, query_vector



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This defines some selector interface used as glue to combine
DataPack/multiPack processors and Pipeline.
"""
from typing import Generic, Iterator, TypeVar, Optional, Union, Dict, Any

import re
import warnings

from forte.common.configuration import Config
from forte.common.configurable import Configurable
from forte.data.base_pack import BasePack
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack

InputPackType = TypeVar("InputPackType", bound=BasePack)
OutputPackType = TypeVar("OutputPackType", bound=BasePack)

__all__ = [
    "Selector",
    "DummySelector",
    "SinglePackSelector",
    "NameMatchSelector",
    "RegexNameMatchSelector",
    "FirstPackSelector",
    "AllPackSelector",
]


class Selector(Generic[InputPackType, OutputPackType], Configurable):
    def __init__(self):
        self.configs: Config = Config({}, {})

    def select(self, pack: InputPackType) -> Iterator[OutputPackType]:
        raise NotImplementedError

    def initialize(
        self, configs: Optional[Union[Config, Dict[str, Any]]] = None
    ):
        self.configs = self.make_configs(configs)


class DummySelector(Selector[InputPackType, InputPackType]):
    r"""Do nothing, return the data pack itself, which can be either
    :class:`~forte.data.data_pack.DataPack` or :class:`~forte.data.multi_pack.MultiPack`.
    """

    def select(self, pack: InputPackType) -> Iterator[InputPackType]:
        yield pack


class SinglePackSelector(Selector[MultiPack, DataPack]):
    """
    This is the base class that select a DataPack from MultiPack.
    """

    def select(self, m_pack: MultiPack) -> Iterator[DataPack]:
        reverse = self.configs.reverse_selection

        for name, pack in m_pack.iter_packs():
            if reverse:
                if not self.will_select(name, pack, m_pack):
                    yield pack
            else:
                if self.will_select(name, pack, m_pack):
                    yield pack

    def will_select(
        self, pack_name: str, pack: DataPack, multi_pack: MultiPack
    ) -> bool:
        """
        Implement this method to return a boolean value whether the
        pack will be selected.

        Args:
            pack_name: The name of the pack to be selected.
            pack: The pack that needed to be determined whether it will be
              selected.
            multi_pack: The original multi pack.

        Returns:
            A boolean value to indicate whether `pack` will be returned.
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {"reverse_selection": False}


class NameMatchSelector(SinglePackSelector):
    r"""
    Select a :class:`~forte.data.data_pack.DataPack` from a
    :class:`~forte.data.multi_pack.MultiPack` with specified
    name. This implementation takes special care for backward compatibility.

    Deprecated:

        .. code-block:: python

            selector = NameMatchSelector(select_name="foo")
            selector = NameMatchSelector("foo")

    Now:

        .. code-block:: python

            selector = NameMatchSelector()
                selector.initialize(
                    configs={
                        "select_name": "foo"
                    }
            )

    WARNING: Passing parameters through __init__ is deprecated, and does not
    work well with pipeline serialization.
    """

    def __init__(self, select_name: Optional[str] = None):
        super().__init__()
        self.select_name = select_name
        if select_name is not None:
            warnings.warn(
                (
                    "Passing parameters through __init__ is deprecated,"
                    " and does not work well with pipeline serialization."
                )
            )

    def will_select(
        self, pack_name: str, pack: DataPack, multi_pack: MultiPack
    ):
        return pack_name == self.select_name

    def initialize(
        self, configs: Optional[Union[Config, Dict[str, Any]]] = None
    ):
        super().initialize(configs)
        try:
            configs_ = configs.todict()  # type:ignore
        except AttributeError:
            configs_ = {} if configs is None else configs

        if self.select_name is not None:
            configs_["select_name"] = self.select_name
        super().initialize(configs_)

        if self.configs["select_name"] is None:
            raise ValueError("select_name shouldn't be None.")
        self.select_name = self.configs["select_name"]

    @classmethod
    def default_configs(cls):
        return {"select_name": None}


class RegexNameMatchSelector(SinglePackSelector):
    r"""Select a :class:`~forte.data.data_pack.DataPack` from a
    :class:`~forte.data.multi_pack.MultiPack` using a regex.

    This implementation takes special care for backward compatibility.

    Deprecated:

        .. code-block:: python

            selector = RegexNameMatchSelector(select_name="^.*\\d$")
            selector = RegexNameMatchSelector("^.*\\d$")

    Now:

        .. code-block:: python

            selector = RegexNameMatchSelector()
            selector.initialize(
                configs={
                    "select_name": "^.*\\d$"
                }
            )

    WARNING:
        Passing parameters through __init__ is deprecated, and does not
        work well with pipeline serialization.

    """

    def __init__(self, select_name: Optional[str] = None):
        super().__init__()
        self.select_name = select_name
        if select_name is not None:
            warnings.warn(
                (
                    "Passing parameters through __init__ is deprecated,"
                    " and does not work well with pipeline serialization."
                )
            )

    def will_select(
        self, pack_name: str, pack: DataPack, multi_pack: MultiPack
    ) -> bool:
        return re.match(self.select_name, pack_name) is not None  # type:ignore

    def initialize(
        self, configs: Optional[Union[Config, Dict[str, Any]]] = None
    ):
        super().initialize(configs)

        try:
            configs_ = configs.todict()  # type:ignore
        except AttributeError:
            configs_ = {} if configs is None else configs

        if self.select_name is not None:
            configs_["select_name"] = self.select_name

        super().initialize(configs_)

        if self.configs["select_name"] is None:
            raise ValueError("select_name shouldn't be None.")
        self.select_name = self.configs["select_name"]

    @classmethod
    def default_configs(cls):
        return {"select_name": None}


class FirstPackSelector(SinglePackSelector):
    r"""Select the first entry from :class:`~forte.data.multi_pack.MultiPack` and yield it."""

    def will_select(
        self, pack_name: str, pack: DataPack, multi_pack: MultiPack
    ) -> bool:
        return multi_pack.pack_names[0] == pack_name


class AllPackSelector(SinglePackSelector):
    r"""Select all the packs from :class:`~forte.data.multi_pack.MultiPack` and yield them."""

    def will_select(
        self, pack_name: str, pack: DataPack, multi_pack: MultiPack
    ) -> bool:
        return True
